#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 17 13:30:07 2023

@author: maria
"""
#### Simulating structurally variable NPCs for microscpy ####
###
import exportCSV
import NPC
import NPC_plotting
import Analyse_deformed
#import utility

# import numpy as np

export = True  # set to True to export data

data_dir = "./data/"  # Directory for output files

config = "config.yaml"
var = NPC.getVars(config)  # Transform config file into a python dictionary

#### Adjust simulation parameters
var["n"] = 7  # Number of NPCs to be simulated
NPCi = (
    0  # index out of n of NPC to be shown in any detail-plot or animation. 0-indexed.
)
var["seed"] = 54648  # seed for reproducibility. Any number but 0

## Select one or more nups, their terminus, and an NPC model for simulation
var["nup"] = ("nup107",)
var["term"] = ("N",)
var["model"] = "5a9q"
# var["rel"] = True # remove the "#" before var["rel"] = True to select the first nup as reference

#### Variability parameters
### Irregular variability
var[
    "mag"
] = 20  # Magnitude of irregular variability. 0: No deformation, 15: Strong deformation
var["zmag"] = var["mag"] / 2  # magnitude of offset in z. Not computed via springs

######################### Geometric variability
# var["symmet"] = 8 # Symmetry of simulated NPCs

## Mean taken from input-model if "None"
# var["rnew"] = None # Mean radius [nm]. Positive number float or int
# var["rsigma"] = None # Standard-deviation radius. . Positive number float or int

# var["dnew"] = None # Mean ring distance [nm]. Positive number float or int
# var["dsigma"] = None # Standard deviation ring distance. Positive number float or int

# var["kappa"] =  None # Controls tilt of individual rings, Kappa of von-mises fisher distribution. Positive number float or int or 0
# var[
#     "shiftsigma"
# ] = None  # Controls shift of individual rings. Standard deviation of 0-centred normal [nm]

# # Twist between nucleoplasmic and cytoplasmic ring
# var["thetanew"] = None # Mean twist angle [rad]
# var["thetasigma"] = None # standard deviation twist angle

# var["elliptnew"] = None # approximate semiminor/semimajor axis ratio.
# var["elliptsigma"] = None # Standard deviation of semiminor/semimajor axis ratio

########################

#### Run simulations
NPCs = NPC.getNPCs(var)  # Dictionary of all simulated NPCs
NPCscoords, offsetNPCs = NPC.getNPCcoords(NPCs, var, offset=True)

#### Visualisation
# Overview
NPC_plotting.plotOverview(offsetNPCs, NPCs, var, width=10)

# NPC_plotting.positionVStime(NPCs, index = NPCi, legend = True)
NPC_plotting.plotDetail(
    NPCscoords, NPCs, var, index=NPCi, width=5, mode="2D", trajectory=False
)
# %matplotlib qt # run to view animation. Default is %matplotlib inline
# name = None #
# Run line below twice if animation doesn't show
# NPC_plotting.AnimateOverview(NPCs, offsetNPCs, var, width = 12, directory = data_dir, name = name, ext = ".gif")

#### Analyse generated NPCs
ellipse_allrings = Analyse_deformed.ellipses(NPCscoords, membership=NPCs["z_i_all"])
ellipse_CRNR = Analyse_deformed.ellipses(NPCscoords, membership=NPCs["ringmemall"])

circle_allrings = Analyse_deformed.circles(NPCscoords, membership=NPCs["z_i_all"])
circle_CRNR = Analyse_deformed.circles(NPCscoords, membership=NPCs["ringmemall"])

# Compute further features
featuresAll = Analyse_deformed.meanfeaturesC(NPCs, var, circle_allrings) # featuresAll: r, sqsum, residual, zsqsum of circle per NPC
featuresElAll = Analyse_deformed.meanfeaturesE(NPCs, var, ellipse_allrings)
featuresel3DAll = exportCSV.colfeaturesEl3D(NPCs, ellipse_CRNR) # featuresel3DAll: centre subcomplex, 1 centre subcomplex 2, centre subcomplex n, tilt subcomplex 1, tilt subcomplex 2 ...

## Show histogram of features
NPC_plotting.gethistdata(var, NPCs, featuresAll, featuresElAll, featuresel3DAll, width = 5, bins = None)

#### Export data
## All NPC coordinates and metadata
if export:
    nameDict, name = exportCSV.names(var)
    exportCSV.MakeCSV(var, NPCs, NPCscoords, nameDict, name, data_dir)

    ## features whole NPC
    featurescsv = exportCSV.featuresCSV(
        NPCs, var, name, circle_allrings, ellipse_allrings, data_dir
    )

    ## features per NPC ring
    exportCSV.featuresCSV_rings(
        NPCs, var, name, data_dir, circle_allrings, ellipse_allrings
    )

    ## features per subcomplex
    exportCSV.featuresCSV_subcomplex(NPCs, circle_CRNR, ellipse_CRNR, name, data_dir)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar  8 11:25:03 2023

@author: maria
"""
import yaml
import DeformNPC
import numpy as np
import math

def getVars(config):
    """reads a .yaml file and turns it into a python dictionary
    
    :param config: yaml file to be read 
    :returns: var, a python dictionary version of the config file, which can 
    be modified 
    """
    with open(config) as file:
        var = dict(yaml.load(file, Loader=yaml.FullLoader))
    return var


def getNPCs(var):
    """
    :param var: dictionary of simulation parameters
    returns: Dictionary containing simulated NPCs and their metadata
    """

    return DeformNPC.MultipleNPC(**var)
    # return DeformNPC.MultipleNPC(var["nup"], var["term"], var["model"], n = var["n"], rel = var["rel"], rvar=var["rvar"],
    #                                  thetavar = var["thetavar"], dvar = var["dvar"], symmet = var["symmet"],
    #                                  elliptvar = var["elliptvar"], mag = var["mag"], zmag = var["zmag"], seed = var["seed"])




def getOffsetNPCs(NPCcoords, offset=None, offsetmult=None, justoffset:bool = False):
    """Arrange NPCs on a grid by offsetting them in x and y direction
    :param NPCcoords: Coordinates of NPC, non-offset
    :param offset: distance by which each NPC should be offset. Automatically determined if not provided
    :param offsetmult: multiplier of offset. Default 1.25
    :param justoffset: Only returns offset in nm if True. Default False 
    :returns:
        offsetNPCs: Offset coordinates"""

    offsetmult = 1 if not offsetmult else offsetmult
    offset = offsetmult * np.max(np.abs(NPCcoords[:, :2])) if not offset else offset
  
    if not justoffset:  
        n = int(NPCcoords[-1, 4] + 1)  # number of NPCs
        offsetNPCs = np.copy(NPCcoords)


        # Determine the number of rows and columns needed. The last cells on the grid might stay empty
        ncols = math.ceil(np.sqrt(n))
        nrows = math.ceil(n / ncols)
    
        x = 0  # indexing x coordinate
        y = 1  # indexing y coordinate
        i = 0  # will get updated
    
        for row in range(ncols):
            for col in range(nrows):
                if i < n:
                    offsetNPCs[np.where(offsetNPCs[:, 4] == i), y] += col * 3 * offset
                    offsetNPCs[np.where(offsetNPCs[:, 4] == i), x] += row * 3 * offset
                    i += 1
    
        return offsetNPCs
    return offset


def getNPCcoords(NPCs, var, offset:bool=False, offsetmult = 1, justoffset:bool = False):
    """Generates coordinates of all NPCs with NPCs and nups indexed. 
    :param NPCs: Dictionary containing simulated NPCs and their metadata
    :param var: Dictionary of simulation parameters
    :param offset: NPCs are arranged on a grid if True. Offset can be applied as 
    a separate function if both offset and non-offset NPCs are required. Defaults to False
    :returns: Coordinates of NPCs or coordinates of NPCs and offset coordinates of NPCs of offet = True
    """
    npcs = NPCs["NPCs"]  # Does not yet include tilt and shift
    nupIndex = NPCs["nupIndex"]

    zoffsets = NPCs["zoffsets"]
    NucSideBool = NPCs["nuclear_side_boolean"]

    tiltnucv, tiltcytv = DeformNPC.tiltvectors(var["kappa"], var["n"], var["seed"])
    shiftNuc, shiftCyt = DeformNPC.shiftvectors(
        var["shiftsigma"], var["n"], var["seed"]
    )


    NPCscoords = DeformNPC.MultipleNPCs_coord(
        npcs,
        zoffsets,
        var["symmet"],
        NucSideBool,
        nupIndex,
        tiltnucv,
        tiltcytv,
        shiftNuc,
        shiftCyt,
        seed=var["seed"],
    )
    
    if justoffset:
        return getOffsetNPCs(NPCscoords, justoffset = justoffset)
    
    if offset: 
        return NPCscoords, getOffsetNPCs(NPCscoords, offsetmult=offsetmult)  
    
    return NPCscoords

    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct  6 16:21:05 2021

@author: maria
"""
import numpy as np
import math
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from mpl_toolkits.mplot3d.art3d import Line3DCollection
import seaborn as sns
from warnings import warn
import matplotlib.animation as animation
import DeformNPC
import Analyse_deformed
from copy import deepcopy
import matplotlib.cm as cm
import matplotlib.colors as colors
from matplotlib.colors import ListedColormap
from matplotlib.lines import Line2D
from scipy.spatial.transform import Rotation as Rot
from sympy.ntheory import primefactors
from scipy.stats import norm
import copy
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
from IPython.display import HTML
import npc

Pos3D = DeformNPC.Pos3D
Sol3D = DeformNPC.Sol3D



class plotOverview:
    def __init__(
        self,
        offsetNPCs,
        NPCs,
        var,
        width=10,
        showforces:bool=False,
        ellipse=False,
        circle=False,
        perRing:bool = False, 
        view:str="front",
        markersizemult=10,
    ):
        """ Plot all NPCs
        :param offsetNPCs: NPC coordinates arranged on a grid 
        :param NPCs: Dictionary containing simulated NPCs and their metadata
        :param var: Dictionary of simulation parameters
        :param width: Width of the figure 
        :param showforces: Show force coordinates, defaults to False
        :param ellipse: Plot fitted ellipses
        :param circles: Plot fitted circles
        :param perRing: Plots circles or ellipses per subcomplex if False, or per Ring if True, defaults to False 
        :param view: Show plot in front- or side view, "front", or "side". Defaults to "front"
        "param markersizemult": Multiplier of plot markersize. Defaults to 10. 
        """
        self.view = view
        self.markersizemult = markersizemult
        if showforces: 
            self.NPCs = NPCs 
            self.var = var
        forcecoords = DeformNPC.Multiple_coord(NPCs["fcoords"], var["symmet"], NPCs["nupIndex"])

        membership = NPCs["z_i_all"] if perRing else NPCs["ringmemall"]

        self.OverviewPlot(
            offsetNPCs,
            forcecoords,
            var["mag"],
            [0],
            membership,
            width,
            showforces=showforces,
            ellipse=ellipse,
            circle=circle,
        )

    def OverviewPlot(
        self,
        offsetNPCs,
        forcecoords,
        mag,
        r,
        membership,
        width,
        showforces=False,
        ellipse=False,
        circle=False,
    ):

        n = int(offsetNPCs[-1, 4] + 1)  # number of NPCs

        if n == 1:
            markersize = self.markersizemult * width
        elif n <= 4:
            markersize = 2 * width
        else:
            markersize = 0.1 * width

        # prepare to colourcode z position
        zs = offsetNPCs[:, 2]
        nup_i = offsetNPCs[:, 3].astype(int)

        zcolour = []
        zcolour.extend(ColourcodeZ(list(zs)))

        # plot
        fig, ax = plt.subplots(1, 1, figsize=(width, width))
        ax.set_title("mag " + str(mag))

        nupmarker = nupMarker()
        nupcolor = nupColor()

        dim2 = 2 if self.view == "side" else 1  # 1 is y-coordinates, 2 is z-coordinates
        for i in np.unique(nup_i):
            x = offsetNPCs[nup_i == i][:, 0]
            y = offsetNPCs[nup_i == i][:, dim2]
            zcolour_subset = [zcolour[j] for j in range(len(zcolour)) if nup_i[j] == i]
            ax.scatter(
                x,
                y,
                c=zcolour_subset,
                s=40 * markersize,
                edgecolors=nupcolor[i],
                marker=nupmarker[i],
            )

        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        ax.spines["left"].set_visible(False)

        if showforces == True:
            offset = npc.getNPCcoords(self.NPCs, self.var, justoffset = True)
            forceoffset = npc.getOffsetNPCs(forcecoords, offset = offset)
            ax.scatter(
                forceoffset[:, 0], forceoffset[:, 1], c="blue", s=2 * markersize, marker = "x"
            )

        # fit circle and/or ellipse
        if ellipse:
            NPCs_ellipse = Analyse_deformed.ellipses(offsetNPCs, membership=membership)
            if type(membership[0]) == np.str_:
                for i in range(n):
                    for key in NPCs_ellipse:
                        el = plotEllipse(NPCs_ellipse[key][i])
                        alpha = [float(i) for i in ColourcodeZ(el[:, 2])]
                        ax.scatter(el[:, 0], el[:, 1], lw=2, c="g", alpha=alpha)

            elif type(membership[0]) == np.int64 or type(
                membership[0] == np.float64
            ):  # if one circle fitted per z-ring
                for i in range(len(NPCs_ellipse)):
                    el = plotEllipse(NPCs_ellipse[i])
                    alpha = [float(i) for i in ColourcodeZ(el[:, 2])]
                    ax.scatter(el[:, 0], el[:, 1], lw=1, c="g", alpha=alpha)

        if circle:
            NPCs_circle = Analyse_deformed.circles(offsetNPCs, membership=membership)

            if (
                type(membership[0]) == np.str_
            ):  # if one circle fitted per NPC subcomplexD
                for i in range(n):
                    for key in NPCs_circle:
                        circ = plotCircle(NPCs_circle[key][i])
                        alpha = [float(i) for i in ColourcodeZ(circ[:, 2])]
                        ax.scatter(circ[:, 0], circ[:, 1], lw=2, c="b", alpha=alpha)

            elif type(membership[0]) == np.int64 or type(
                membership[0] == np.float64
            ):  # if one circle fitted per z-ring
                for i in range(len(NPCs_circle)):
                    circ = plotCircle(NPCs_circle[i])

                    alpha = [float(i) for i in ColourcodeZ(circ[:, 2])]
                    if math.isnan(alpha[0]):
                        alpha = 0.7

                    ax.scatter(circ[:, 0], circ[:, 1], lw=1, c="b", alpha=alpha)

        # colourbar for z position
        mincolour = float(min(zcolour))
        maxcolour = float(max(zcolour))

        if mincolour != maxcolour:
            incr = (maxcolour - mincolour) / max(zs)  # increment in colour
            cmap = ListedColormap(
                [
                    str(i)
                    for i in list(np.arange(mincolour, maxcolour + 0.5 * incr, incr))
                ]
            )
            norm = plt.Normalize(min(zs), max(zs))
            fig.colorbar(
                cm.ScalarMappable(norm=norm, cmap=cmap),
                shrink=0.7,
                label="z [nm]",
                ticks=[min(zs), max(zs)],
            )

        ax.set_xticks([])
        ax.set_yticks([])

        barsize = 20
        label = str(barsize) + " nm"
        # label = ""
        scalebar = AnchoredSizeBar(
            ax.transData,
            barsize,
            label,
            "lower right",
            size_vertical=5,
            # bbox_to_anchor = (1050, 70), frameon=False,
            # bbox_to_anchor = (1050, 70),
            frameon=False,
            color=[0.3, 0.3, 0.3],
        )
        ax.add_artist(scalebar)

        ax.axis("scaled")
        fig.tight_layout()


def plotEllipse(NPC):
    return generate_ellipse_by_vectors(NPC["Ce"], NPC["minv"], NPC["majv"])


def plotCircle(NPC):
    (
        r,
        _,
        _,
        _,
        C,
        _,
        normal1,
        normal0,
    ) = NPC  # radius, centre, vectors collinear with circle-plane
    return generate_circle_by_vectors(C, r, normal0, normal1)


def generate_circle_by_vectors(C, r, normal0, normal1):
    t = np.linspace(0, 2 * np.pi, 150)
    P_circle = (
        r * np.cos(t)[:, np.newaxis] * normal0
        + r * np.sin(t)[:, np.newaxis] * normal1
        + C
    )
    return P_circle


def generate_ellipse_by_vectors(C, minv, majv):
    t = np.linspace(0, 2 * np.pi, 100)
    P_ellipse = np.cos(t)[:, np.newaxis] * minv + np.sin(t)[:, np.newaxis] * majv + C
    return P_ellipse


def ColourcodeZ(z, darkest=0.2, brightest=0.7):
    """colourcode z, smaller values are shown darker"""
    return [str(i) for i in np.interp(z, (min(z), max(z)), (darkest, brightest))]


def shifttilt_all_t(pos3D, nFrames, centre, tilt):
    pos3Dc = copy.deepcopy(pos3D)

    if type(tilt) is not type(None):
        for t in range(nFrames):  # tilt for all timeframes
            tiltrot = Rot.from_rotvec(tilt)  #
            pos3Dc[t] = tiltrot.apply(pos3Dc[t])

    for t in range(nFrames):  # shift for all timeframes
        pos3Dc[t, :, 0] += centre[0]
        pos3Dc[t, :, 1] += centre[1]

    return pos3Dc


def plotDetail(
    NPCscoords,
    NPCs:dict,
    var:dict,
    index:int=0,
    width=4,
    mode:str="3D",
    showforces:bool=False,
    trajectory:bool=False,
):
    """ Generate a detail plot of an NPC with given index
    :param NPCscoords: Coordinates of all NPCs
    :param NPCs: Dictionary containing simulated NPCs and their metadata
    :param var: Dictionary of simulation parameters
    :param index: Index of NPC for which the plot should be generated
    :param mode: "3D" or "2D", default: 3D: 2D or 3D plot
    :type mode: str
    :param showforces: Show deforming forces. Will only be correct if tilt and shift have not been applied
    :param trajectories: Show trajectories of nodes over time. Will only be correct if tilt and shift have not been applied
    """

    NoneOrTilt = lambda a, i: a[i] if type(a) is not type(None) else None

    plotNPC = NPCscoords[NPCscoords[:, 4] == index]
    tiltnucv, tiltcytv = DeformNPC.tiltvectors(var["kappa"], var["n"], var["seed"])
    shiftNuc, shiftCyt = DeformNPC.shiftvectors(
        var["shiftsigma"], var["n"], var["seed"]
    )
    tiltnucv_i = NoneOrTilt(
        tiltnucv, index
    )  # tiltnucv[NPCi] if type(tiltnucv) is not type(None) else None
    tiltcytv_i = NoneOrTilt(tiltcytv, index)

    if mode == "2D":
        Plot2D(
            plotNPC,
            NPCs["NPCs"][index],
            NPCs["zexp"],
            NPCs["nupIndex"][index],
            var["symmet"],
            var["nConnect"],
            tiltnucv_i,
            tiltcytv_i,
            shiftNuc[index],
            shiftCyt[index],
            forces=NPCs["fcoords"][index],
            showforces=showforces,
            legend=False,
            trajectory=trajectory,
            colourcode=True,
            springs=False,
            anchorsprings=False,
            width=width,
        )

    elif mode == "3D":
        membership = NPCs["ringmemall"][NPCscoords[:, 4] == index]
        tiltnucv_i = NoneOrTilt(
            tiltnucv, index
        )  # tiltnucv[NPCi] if type(tiltnucv) is not type(None) else None
        tiltcytv_i = NoneOrTilt(tiltcytv, index)
        Plot3D(
            plotNPC,
            NPCs["NPCs"][index],
            NPCs["nupIndex"][index],
            var["nup"],
            var["term"],
            NPCs["isnuc"],
            var["symmet"],
            tiltnucv_i,
            tiltcytv_i,
            shiftNuc[index],
            shiftCyt[index],
            NPCs["fmags"][index],
            NPCs["fcoords"][index],
            membership,
            showforces=showforces,
            trajectory=trajectory,
            viewFrame=-1,
            width=width,
        )


def nupMarker():
    return ["o", "s", "p", "*", "X", "^"]


def nupColor():
    return ["gray", "cyan", "magenta", "green", "purple", "blue"]


def Plot2D(
    plotNPC,
    solution,
    z,
    nup_i,
    symmet,
    nConnect,
    tiltnucv,
    tiltcytv,
    shiftNuc,
    shiftCyt,
    linestyle="-",
    trajectory=True,
    colourcode=True,
    springs=False,
    anchorsprings=False,
    markersize=20,
    forces=0,
    showforces=False,
    legend=False,
    width=5,
):
    """
    solution: Output of solve_ivp
    symmet: number of nodes per ring
    nConnect: number of neighbours connected on each side per node
    linestyle (default: "-"): Linestyle in 1st plot
    legend (default: False): Whether to show a legend in the 1st plot
    colourcode (default: True): colourcodes trajectory with velocity
    colourbar (default: True): Plots colourbar in 2nd plot if True and if colourcode is True
    mainmarkercolor: Colour of nodes in 2nd plot
    """

    nRings = len(z)
    # plt.rcParams.update({'font.size': 25})

    fig, ax = plt.subplots(1, 1, figsize=(width, width))  # 16, 16
    viewFrame = -1  # 0 is the first frame, -1 is the last frame
    mainmarkercolor = ColourcodeZ(z)
    marker = nupMarker()
    edgecolor = nupColor()

    for i in range(nRings):
        centre = shiftNuc if z[i] < np.mean(z) else shiftCyt

        nFrames = len(solution[i].t)  # Nodes at last timestep
        pos3D, vel3D = copy.deepcopy(Sol3D(solution[i]))

        tilt = tiltnucv if z[i] < np.mean(z) else tiltcytv
        pos3D = shifttilt_all_t(pos3D, nFrames, centre, tilt)
        nup = int(nup_i[i])

        ax.plot(
            pos3D[viewFrame, :symmet, 0],
            pos3D[viewFrame, :symmet, 1],
            linestyle="",
            marker=marker[nup],
            color=edgecolor[nup],
            markerfacecolor=mainmarkercolor[i],
            markersize=markersize,
            zorder=50,
            label=str(round(z[i], 1)),
        )

        if anchorsprings:
            ax.plot(centre[0], centre[1], marker="o", color="gray", markersize=15)
            for j in range(symmet):
                ax.plot(
                    (pos3D[viewFrame, j, 0], centre[0]),
                    (pos3D[viewFrame, j, 1], centre[1]),
                    linestyle="-",
                    marker="",
                    color="lightgray",
                    zorder=2,
                    alpha=0.5,
                )

        # circumferential springs
        if springs:
            for ni in range(1, nConnect + 1):  # neighbours to connect to
                for j in range(symmet):  # node to connect from
                    ax.plot(
                        pos3D[viewFrame, (j, (j + ni) % symmet), 0],
                        pos3D[viewFrame, (j, (j + ni) % symmet), 1],
                        linestyle="-",
                        marker="",
                        color="lightgray",
                        zorder=3,
                        alpha=0.5,
                    )  # , linewidth = 5)

        if trajectory:
            if colourcode:  # Colourcoded trajectory
                ### colourcoding velocities
                normvel = np.zeros((nFrames, symmet))  # nFrames, node

                for j in range(symmet):
                    for frame in range(nFrames):
                        normvel[frame, j] = np.linalg.norm(
                            [vel3D[frame, j, 0], vel3D[frame, j, 1], vel3D[frame, j, 2]]
                        )

                norm = plt.Normalize(normvel.min(), normvel.max())

                #####trajectory colorcoded for velocity
                for j in range(symmet):
                    points = pos3D[:, j, :].reshape(-1, 1, 3)  # 3D done
                    segments = np.concatenate([points[:-1], points[1:]], axis=1)

                    lc = LineCollection(
                        segments[:, :, :2],
                        cmap="plasma",
                        norm=norm,
                        zorder=4,
                        linewidth=3.0,
                    )
                    lc.set_array(normvel[:, j])
                    line = ax.add_collection(lc)

            else:  # monochrome trajectory
                for j in range(symmet):
                    ax.plot(pos3D[:, j, 0], pos3D[:, j, 1], color="blue", linestyle="-")

        ### Force vectors
        if showforces and type(forces) != int:
            forces2d = forces[i]
            for j in range(symmet):
                ax.arrow(
                    x=pos3D[0, j, 0],  # frame, node, x or y
                    y=pos3D[0, j, 1],
                    dx=(forces2d[j, 0] - pos3D[0, j, 0]),
                    dy=(forces2d[j, 1] - pos3D[0, j, 1]),
                    width=0.75,
                    color="blue",
                    zorder=101,
                    alpha=1,
                )

    if trajectory and colourcode and legend:
        # fig.legend(bbox_to_anchor=(0.1,-0.025), loc="lower left", ncol = 4, title = "z [nm]")#loc="best")
        axcb = fig.colorbar(line, ax=ax, shrink=0.7, aspect=50, ticks=[0], pad=0.01)
        axcb.set_label("velocity (a.u.)")

    legend_elements = [
        Line2D(
            [0],
            [0],
            marker="o",
            linestyle="",
            color="0.7",
            label="CR",
            markerfacecolor="0.7",
            markeredgecolor="0.7",
            mew=6,
            markersize=20,
        ),
        Line2D(
            [0],
            [0],
            marker="o",
            linestyle="",
            color="0.2",
            label="NR",
            mew=6,
            markersize=20,
        ),
        Line2D(
            [0],
            [0],
            marker="$\\rightarrow$",
            linestyle="",
            color="blue",
            label="force",
            markersize=30,
        ),
        Line2D(
            [0],
            [0],
            marker="",
            linestyle="-",
            color="lightgray",
            label="spring",
            alpha=0.5,
        ),
    ]  # Line2D([0], [0], marker='', linestyle="-", color='m', label='fit', linewidth=2)]

    ax.legend(
        handles=legend_elements,
        loc="upper right",
        bbox_to_anchor=(0.9, -0.1),
        ncol=4,
        fontsize=24,
        frameon=False,
    )

    ax.axis("scaled")
    ax.set(xlabel="x (nm)", ylabel="y (nm)")
    plt.tight_layout()


def positionVStime(NPCs, index=0, width=3, legend=False):
    """
    NPCs: Dictionary containing info of all NPCs
    index: Index of NPC for which plot should be generated
    width: Width of each subplot
    legend: Show ledged. Default False
    """
    symmet = len(
        NPCs["fmags"][index][0]
    )  # length of array of deforming forces on NPC i ring 0. Should be equivalent to symmetry
    XYoverTime(NPCs["NPCs"][index], symmet=symmet, width=width, legend=legend)


def XYoverTime(solution, symmet, width=5, legend=False):
    """x and y positions over time"""

    nRings = len(solution)

    # determin number of rows and colums in final figure. One plot per NPC ring
    l = 2 - nRings % 2
    rows, cols = sorted((int((nRings / l)), l))

    fig, ax = plt.subplots(rows, cols, figsize=(width * rows, width * cols))
    palette = sns.color_palette("hsv", 2 * symmet)

    for ring in range(nRings):
        for i in range(symmet):
            if nRings > 1:
                ax = ax.flatten()
            labelx = labely = None
            if ring == 0:
                labelx = "x" + str(i)
                labely = "y" + str(i)
                labelz = "z" + str(i)

            if nRings > 1:
                ax[ring].plot(
                    solution[ring].t,
                    Pos3D(solution[ring])[:, i, 0],
                    label=labelx,
                    linestyle="-",
                    color=palette[i * 2],
                )
                ax[ring].plot(
                    solution[ring].t,
                    Pos3D(solution[ring])[:, i, 1],
                    label=labely,
                    linestyle="--",
                    color=palette[i * 2],
                )
                ax[ring].plot(
                    solution[ring].t,
                    Pos3D(solution[ring])[:, i, 2],
                    label=labelz,
                    linestyle="-.",
                    color=palette[i * 2],
                )
                ax[ring].set_title("ring " + str(ring))
                ax[ring].set(xlabel="t (a.u.)")
                ax[ring].set(ylabel="change in x or y [nm]")

            else:
                ax.plot(
                    solution[ring].t,
                    Pos3D(solution[ring])[:, i, 0],
                    label=labelx,
                    linestyle="-",
                    color=palette[i * 2],
                )
                ax.plot(
                    solution[ring].t,
                    Pos3D(solution[ring])[:, i, 1],
                    label=labely,
                    linestyle="--",
                    color=palette[i * 2],
                )
                ax.plot(
                    solution[ring].t,
                    Pos3D(solution[ring])[:, i, 2],
                    label=labelz,
                    linestyle="-.",
                    color=palette[i * 2],
                )
                ax.set_title("ring " + str(ring))
                ax.set(xlabel="t (a.u.)")
                ax.set(ylabel="change in x or y [nm]")

    if legend:
        fig.legend(bbox_to_anchor=(1, 0.9), loc="upper left")
    plt.tight_layout()

    plt.show()


def Plot3D(
    plotNPC,
    solution,
    nup_i,
    nupname,
    termname,
    isnuc,
    symmet,
    tiltnucv,
    tiltcytv,
    shiftNuc,
    shiftCyt,
    fmags,
    fcoords,
    membership,
    showforces=False,
    trajectory=False,
    viewFrame=-1,
    colour=["black", "gray"],
    width=5,
):
    """viewFrame: 0 is first frame, -1 is last frame"""
    fig = plt.figure(figsize=(width, width))
    ax = fig.add_subplot(111, projection="3d")

    linewidth = 3
    nRings = len(nup_i)

    marker = nupMarker()
    color = nupColor()

    zs = plotNPC[:, 2]
    colour = ColourcodeZ(zs)

    nupold = None
    for ring in range(nRings):
        # ax.scatter(Pos3D(solution[ring])[viewFrame, : ,0], Pos3D(solution[ring])[viewFrame, :,1], Pos3D(solution[ring])[viewFrame, :,2], s = 300, c = str(colour[ring]), linewidths = linewidth,  marker = "o")

        x = plotNPC[ring * symmet : (ring + 1) * symmet, 0]
        y = plotNPC[ring * symmet : (ring + 1) * symmet, 1]
        z = zs[ring * symmet : (ring + 1) * symmet]

        nup = int(nup_i[ring])
        if ring > 0:
            nupold = int(nup_i[ring - 1])
        label = nupname[nup] + " " + termname[nup] if nup != nupold else ""

        ax.scatter(
            x,
            y,
            z,
            s=300,
            c=colour[ring * symmet : (ring + 1) * symmet],
            edgecolors=color[nup],
            linewidths=linewidth,
            marker=marker[nup],
            label=label,
        )
    ax.legend()

    if showforces:
        for ring in range(nRings):
            x = plotNPC[ring * symmet : (ring + 1) * symmet, 0]
            y = plotNPC[ring * symmet : (ring + 1) * symmet, 1]
            z = zs[ring * symmet : (ring + 1) * symmet]

            for node in range(symmet):
                ax.quiver(
                    x[node],
                    y[node],
                    z[node],
                    fcoords[ring][node][0],
                    fcoords[ring][node][1],
                    0,
                    length=fmags[ring][node],
                    normalize=True,
                    linewidth=linewidth,
                    edgecolor="blue",
                )

    ax.set_xlabel("x [nm]", labelpad=30)
    ax.set_ylabel("y [nm]", labelpad=30)
    ax.set_zlabel("z [nm]", labelpad=40, fontsize=40)

    ax.set_xticks([-25, 25])
    ax.set_yticks([-25, 25])
    ax.set_zticks([0, 50])

    ax.set_xlim3d(-60, 60)
    ax.set_ylim3d(-60, 60)
    ax.set_zlim3d(0, 100)

    ax.xaxis.pane.fill = False
    ax.yaxis.pane.fill = False
    ax.zaxis.pane.fill = False

    # Trajectories
    if (
        tiltcytv is not None
        or tiltnucv is not None
        or np.any(shiftCyt)
        or np.any(shiftNuc)
    ):
        trajectory = False
    else:
        trajectory = trajectory

    if trajectory:
        ### colourcoding velocities
        for i in range(len(solution)):  # equals number of rings
            nFrames = len(solution[i].t)  # for NPC 0
            normvel = np.zeros((nFrames, symmet))  # nFrames, node

            pos3D, vel3D = copy.deepcopy(Sol3D(solution[i]))  # for NPC 0

            centre = shiftNuc if isnuc[i] else shiftCyt
            tilt = tiltnucv if isnuc[i] else tiltcytv

            pos3D = shifttilt_all_t(pos3D, nFrames, centre, tilt)

            for j in range(symmet):
                for frame in range(nFrames):
                    normvel[frame, j] = np.linalg.norm(
                        [vel3D[frame, j, 0], vel3D[frame, j, 1], vel3D[frame, j, 2]]
                    )

            norm = plt.Normalize(normvel.min(), normvel.max())

            #####trajectory colorcoded for velocity
            for j in range(symmet):
                points = pos3D[:, j, :].reshape(-1, 1, 3)
                segments = np.concatenate([points[:-1], points[1:]], axis=1)
                lc = Line3DCollection(
                    segments, cmap="plasma", norm=norm, zorder=4, linewidth=3.0
                )
                lc.set_array(normvel[:, j])
                line = ax.add_collection(lc)

        axcb = fig.colorbar(line, ax=ax, shrink=0.7, aspect=50, ticks=[0], pad=0.01)
        axcb.set_label("velocity (a.u.)")

    # color of edges that aren't axes
    ax.xaxis.pane.set_edgecolor("w")
    ax.yaxis.pane.set_edgecolor("w")
    ax.zaxis.pane.set_edgecolor("w")
    ax.set_box_aspect([1, 1, 1])
    ax.grid(False)

    fitellipse = False
    if fitellipse:
        NPC_ellipse = Analyse_deformed.ellipses(plotNPC, membership=membership)
        if type(membership[0]) == np.str_:
            for key in NPC_ellipse:
                el = plotEllipse(NPC_ellipse[key][0])
                ax.plot(el[:, 0], el[:, 1], el[:, 2], "g-", lw=4)

        elif type(membership[0]) == np.int64 or type(membership[0] == np.float64):
            for i in range(len(NPC_ellipse)):
                el = plotEllipse(NPC_ellipse[i])
                ax.plot(el[:, 0], el[:, 1], el[:, 2], "g-", lw=4)

    fitcircle = False
    if fitcircle:
        NPCs_circle = Analyse_deformed.circles(plotNPC, membership=membership)

        if type(membership[0]) == np.str_:
            for key in NPCs_circle:
                circ = plotCircle(NPCs_circle[key][0])
                ax.plot(circ[:, 0], circ[:, 1], circ[:, 2], "b-", lw=4)

        elif type(membership[0]) == np.int64 or type(membership[0] == np.float64):
            for i in range(len(NPCs_circle)):
                circ = plotCircle(NPCs_circle[i])
                ax.plot(circ[:, 0], circ[:, 1], circ[:, 2], "b-", lw=4)

    plt.show()


# %matplotlib qt


def AnimateDetail(NPCs, var, index:int=0, width=8, directory="", name=None, ext=".gif"):
    """ Animate an individual structure 
    :param NPCs: Dictionary containing simulated NPCs and their metadata
    :param var: Dictionary of simulation parameters
    :param index: Index of structure to be animated, default 0. 0-indexed 
    :param width: Width of animation plot 
    :param directory: directory to save animation in 
    :param name: name of animation. No file will be saved if None. Defaults to None 
    :param ext: File extension, ".gif" or ".mp4", defaults to ".gif"
    """
    AnimatedScatter(
        NPCs["NPCs"][index],
        var["nConnect"],
        var["symmet"],
        NPCs["rexp"],
        NPCs["fmags"][index],
        NPCs["zexp"],
        NPCs["nupIndex"][index],
        width,
        directory,
        name,
        ext,
    )


class AnimatedScatter(object):


    def __init__(
        self,
        solution,
        nConnect,
        symmet,
        r,
        fmags,
        z,
        nup_i,
        width,
        directory,
        name,
        ext,
    ):
        """An animated scatter plot using matplotlib.animations.FuncAnimation."""
        self.solution = solution
        self.nRings = len(solution)
        self.anchors = np.zeros(3)  # anchors
        self.zcolour = ColourcodeZ(z)
        self.nup_i = nup_i
        framenumbers = []

        for i in range(self.nRings):  # check framenumbers are consistent for each ring
            framenumbers.append(len(self.solution[i].t))
        if len(set(framenumbers)) != 1:
            warn(
                "Number of timesteps for all ring must be the same in order to animate deformation."
            )
            return

        nframes = len(self.solution[0].t)

        self.nConnect = nConnect
        self.symmet = symmet
        self.xy = self.xydata()
        self.stream = self.data_stream(self.xy)

        # Setup the figure and axes...
        self.axscale = 1.25 * (np.amax(fmags) + max(r))
        self.fig, self.ax = plt.subplots(figsize=(width, width))
        # plt.rcParams.update({'font.size': 20})

        # Then setup FuncAnimation.

        if ext == "HTML":
            HTML(self.ani.to_html5_video())

        # if name: self.ani.save(name + ".mp4", dpi = 250, fps = 50)
        if name:
            self.ani = animation.FuncAnimation(
                self.fig,
                self.update,
                interval=(5000 / nframes),
                init_func=self.setup_plot,
                blit=True,
                save_count=len(solution[0].t),
            )
            self.ani.save(directory + name + ext, fps=50, dpi=80)

        else:
            self.ani = animation.FuncAnimation(
                self.fig,
                self.update,
                interval=(5000 / nframes),
                init_func=self.setup_plot,
                blit=True,
                save_count=len(solution[0].t),
            )

        plt.show()

    def xydata(self):
        xy = []
        for ring in range(self.nRings):
            xy.append(
                Pos3D(self.solution[ring])[:, np.append(np.arange(self.symmet), 0)]
            )

        return xy  # [xy[0][:,:,:2]] # back to 2D

    def setup_plot(self):
        """Initial drawing of the scatter plot."""

        self.lines = []
        nupcolors = nupColor()
        nupmarker = nupMarker()

        for i in range(
            int(
                self.nRings
                + self.nRings * self.symmet
                + self.symmet * self.nRings * self.nConnect
            )
        ):
            if i < self.nRings:  # all ring-nodes
                self.lobj = self.ax.plot(
                    [],
                    [],
                    marker=nupmarker[self.nup_i[i]],
                    color=self.zcolour[
                        i
                    ],  # str(np.mean(self.xy[i][0][:,2])), # does not accept color
                    markeredgecolor=nupcolors[self.nup_i[i]],
                    linestyle="",
                    markersize=20,
                )
            elif (
                i >= self.nRings and i <= self.nRings * self.symmet
            ):  # ring(s) to anchor
                self.lobj = self.ax.plot(
                    [], [], marker="", color="orange", linestyle="", alpha=0.1, zorder=0
                )  # anchor
            else:  # circumferential springs
                self.lobj = self.ax.plot(
                    [],
                    [],
                    marker="",
                    color="gray",
                    alpha=0.1,
                    linewidth=4,
                    linestyle="-",
                )

            self.lines.append(self.lobj)

        self.ax.axis("scaled")
        self.ax.set(xlabel="x (nm)", ylabel="y (nm)")
        self.ax.axis([-self.axscale, self.axscale, -self.axscale, self.axscale])

        return [
            self.lines[i][0]
            for i in range(
                int(self.nRings * 2 + self.symmet * self.nRings * self.nConnect)
            )
        ]

    def data_stream(self, pos):
        x = np.zeros((self.symmet + 1, self.nRings))
        y = np.zeros((self.symmet + 1, self.nRings))

        while True:
            for i in range(len(self.xy[0])):  # number of timesteps
                for ring in range(self.nRings):
                    x[:, ring] = self.xy[ring][i][
                        :, 0
                    ]  # x coordinates of ring at timestep i
                    y[:, ring] = self.xy[ring][i][:, 1]
                yield x, y

    def update(self, i):
        """Update the plot."""

        x, y = next(self.stream)

        # Alternating anchor coordinates and x or y coordinates
        xa = np.zeros((2 * self.symmet, self.nRings))
        ya = np.zeros((2 * self.symmet, self.nRings))

        for ring in range(self.nRings):  # alternating coordinates of anchor and node
            for i in range(1, 2 * self.symmet, 2):
                # i = 1,3,5,7,9,11,13,15
                xa[i, ring] = x[int((i - 1) / 2), ring]
                ya[i, ring] = y[int((i - 1) / 2), ring]

        xlist = list(x.T) + list(
            xa.T
        )  # 1st 9 entries are nodes 0-7 and 0, following entries are xa
        ylist = list(y.T) + list(ya.T)

        for i in range(self.nRings):
            self.lines[i][0].set_data(xlist[i], ylist[i])  # i is 0, 1, 2, 3

        for i in range(self.nRings, self.nRings * self.symmet + 1):  # to anchor?
            self.lines[i][0].set_data(
                xlist[2 * (i - 1) : 2 * (i - 1) + 2],
                ylist[2 * (i - 1) : 2 * (i - 1) + 2],
            )

        count = (
            self.nRings + self.nRings * self.symmet
        )  # len(xlist) # for one nRIngs == 1, len(xlist) is 2
        for lnum in range(self.nRings):
            for ni in range(1, self.nConnect + 1):  # neighbours to connect to
                for i in range(self.symmet):  # node to connect from
                    self.lines[count][0].set_data(
                        (xlist[lnum][i], xlist[lnum][(i + ni) % self.symmet]),
                        (ylist[lnum][i], ylist[lnum][(i + ni) % self.symmet]),
                    )
                    count += 1
        # return [self.lines[i][0] for i in range(int(self.nRings*2 + self.symmet*self.nRings*self.nConnect))]
        return [
            self.lines[i][0]
            for i in range(
                int(
                    self.nRings
                    + self.nRings * self.symmet
                    + self.symmet * self.nRings * self.nConnect
                )
            )
        ]


if __name__ == "__main__":
    # a = AnimatedScatter(solution, nRings, nConnect, symmet, r, fmags)

    plt.show()


def AnimateOverview(
    NPCs, offsetNPCs, var, width=8, directory="", name=None, ext=".gif"
):
    """An animated scatter plot using matplotlib.animations.FuncAnimation.
    :param NPCs: Dictionary containing simulated NPCs and their metadata
    :param offsetNPCs: NPC coordinates arranged on a grid 
    :param var: Dictionary of simulation parameters
    :param width: width of animation plot 
    :param directory: directory to save animation in, defaults to "". 
    :param name: name of animation. No file will be saved if None. Defaults to None 
    :param ext: File extension, ".gif" or ".mp4", defaults to ".gif"
    """
    AnimateAll(
        NPCs["NPCs"],
        offsetNPCs,
        NPCs["fcoords"],
        var["symmet"],
        NPCs["rexp"],
        width,
        directory,
        name,
        ext,
    )


class AnimateAll(object):
    def __init__(
        self, NPCs, offsetNPCs, forcecoords, symmet, r, width, directory, name, ext
    ):
        self.n = len(NPCs)
        self.tmax = len(NPCs[0][0].t)
        self.t_all = NPCs[0][0].t
        self.maxr = max(r)
        # self.NPCs = NPCs
        self.symmet = symmet
        self.nRings = len(NPCs[0])
        self.nup_i_all = list(
            offsetNPCs[:, 3]
        )  # indicates which nup each point belongs to

        # self.anccoords = deepcopy(anccoords)
        self.forcecoords = deepcopy(forcecoords)

        NPCsCopy = deepcopy(NPCs)

        # Determine the number of rows and columns needed. The last cells on the grid might stay empty
        self.ncols = math.ceil(np.sqrt(self.n))
        self.nrows = math.ceil(self.n / self.ncols)

        self.nrows = 1 if self.n == 1 else max(primefactors(self.n))
        self.ncols = (
            1 if self.n == 1 else int(self.n / max(primefactors(self.n)))
        )  # math.ceil(self.n/self.ncols)

        # i = 0 # will get updated
        # for row in range(self.ncols):
        #     for col in range(self.nrows):
        #         if (i < self.n):

        #             # self.anccoords[i*self.symmet : (i+1)*self.symmet, 0] += col * 8 * self.maxr
        #             # self.anccoords[i*self.symmet : (i+1)*self.symmet, 1] += row * 8 * self.maxr

        #             self.forcecoords[i*self.symmet : (i+1)*self.symmet, 0] += col * 8 * self.maxr
        #             self.forcecoords[i*self.symmet : (i+1)*self.symmet, 1] += row * 8 * self.maxr
        #             i += 1

        self.xy = self.xydata(NPCsCopy)
        self.stream = self.data_stream(self.xy)

        # Setup the figure and axes...
        # self.axscale = 70
        marginm = 1
        self.xmin = round(min(self.xy[0][:, 0])) - marginm * self.maxr
        self.ymin = round(min(self.xy[0][:, 1])) - marginm * self.maxr
        self.xscale = round(max(self.xy[0][:, 0])) + marginm * self.maxr
        self.yscale = round(max(self.xy[0][:, 1])) + marginm * self.maxr

        self.fig, self.ax = plt.subplots(figsize=(width, width))

        self.ani = animation.FuncAnimation(
            self.fig,
            self.update,
            interval=(5000 / self.tmax),
            init_func=self.setup_plot,
            blit=True,
            save_count=self.tmax,
        )

        # if name: self.ani.save(name + ".mp4", dpi = 250, fps = 30)
        if name:
            self.ani.save(directory + name + ext, dpi=80, fps=30)
        if ext == "HTML":
            HTML(self.ani.to_html5_video())
        else:
            plt.show()

    def xydata(self, NPCsCopy):
        xy0 = []

        xy = []
        for npc in range(self.n):
            for ring in range(self.nRings):
                xy0.append(
                    Pos3D(NPCsCopy[npc][ring])
                )  # [:, np.append(np.arange(self.symmet), 0)])

        for frame in range(self.tmax):
            i = 0  # will get updated.
            for col in range(self.ncols):
                for row in range(self.nrows):
                    if i < self.n * self.nRings:
                        for ring in range(self.nRings):
                            xy0[i][frame][:, 0] += col * 3 * self.maxr  # x
                            xy0[i][frame][:, 1] += row * 3 * self.maxr  # y
                            i += 1

        for frame in range(self.tmax):
            xy.append(
                np.reshape(
                    np.array(xy0)[:, frame, :, :],
                    (self.symmet * self.n * self.nRings, 3),
                )
            )

        return xy

    def setup_plot(self):
        # anchors and forces (remains static)

        x, y, t, c = next(self.stream)
        nupcolor = nupColor()
        ec = [
            nupcolor[int(i)] for i in self.nup_i_all
        ]  # edgecolor for points in plot, coded by nup
        # self.ax.scatter(self.anccoords[:,0], self.anccoords[:,1], s = 5, alpha = 0.1, c = "orange")
        # self.ax.scatter(self.forcecoords[:,0], self.forcecoords[:,1], s = 5, alpha = 0.1, c = "cyan")

        # to be updated

        def truncate_colormap(cmap, minval=1, maxval=0.4, n=4):
            new_cmap = colors.LinearSegmentedColormap.from_list(
                "trunc({n},{a:.2f},{b:.2f})".format(n=cmap.name, a=minval, b=maxval),
                cmap(np.linspace(minval, maxval, n)),
            )
            return new_cmap

        new_cmap = truncate_colormap(plt.get_cmap("Greys"), n=2 * self.nRings)

        self.scat = self.ax.scatter(
            x,
            y,
            marker="o",
            c=c,
            edgecolor=ec,  # np.random.uniform(0, 1, len(c)),
            s=75,
            label="test",
            cmap=new_cmap,
        )

        self.time_text = self.ax.text(0.02, 0.95, "", transform=self.ax.transAxes)

        self.ax.axis("scaled")
        self.ax.set(xlabel="x (nm)", ylabel="y (nm)")
        self.ax.axis([self.xmin, self.xscale, self.ymin, self.yscale])
        self.time_text.set_text("")
        return self.scat, self.time_text

    def data_stream(self, pos):
        # x = np.zeros(self.symmet * self.n) # why not *nRings?
        # y = np.zeros(self.symmet * self.n)

        while True:
            for frame in range(self.tmax):  # number of timesteps
                #  for ring in range(self.nRings):
                x = self.xy[frame][:, 0]  # x coordinates of ring at timestep i
                y = self.xy[frame][:, 1]
                c = self.colourcodeZ(self.xy[frame][:, 2])
                t = self.t_all[frame]
                # c = list(np.repeat("0.5", len(self.xy[t][:, 0])))#list(str(self.xy[t][:, 2]))
                yield x, y, t, c

    def update(self, i):
        """Update the plot."""

        x, y, t, c = next(self.stream)
        self.scat.set_offsets((np.vstack((x, y)).T))
        self.scat.set_array(np.array(c))
        # roundto = 10
        self.time_text.set_text("t: " + str(int(t)))  # str(roundto * round(t/roundto)))

        return self.scat, self.time_text

    def colourcodeZ(self, z, darkest=0.0, brightest=1):
        """colourcode z, smaller values are shown darker, returns array"""
        return np.interp(z, (min(z), max(z)), (darkest, brightest))


class gethistdata:
    def __init__(
        self, var, NPCs, featuresAll, featuresElAll, featuresel3DAll, width=5, bins=None
    ):
        """
        Plots the distribution of different features for all NPCs. Features are 
        NPC radius, minor/major axis ratio of an ellipse, residual sum of squares of a ring-wise fitted ellipse, 
        difference in tilt in radians between the nucleoplasmic and cytoplasmic rings, determined using ellipses to the respective subcomplexes. 
        shift of nucleoplasmic and cytoplasmic ring, determined using ellipses fitted to the respective subcomplexes. 
        Distance between the centroid of the nucleoplasmic and cytoplasmic ring, determined using ellipses fitted to the respective subcomplexes. 
        :param var: Dictionary of simulation parameters
        :param NPCs: Dictionary containing simulated NPCs and their metadata
        :param featuresAll: features of the NPC determined by fitting circles. From Analyse_deformed.meanfeaturesC()
        :param featuresElAll: features of the NPC determined by fitting ellipses. From Analyse_deformed.meanfeaturesE()
        :param featuresel3DAll: centre and tilt of NPC subcomplexes determined by fitting ellipses. from exportCSV.colfeaturesEl3D. 
        :param width: Width of plot, defaults to 5 
        :param bins: Number of bins to be plotted. If None, the number of bins will be automatically determined. Defaults to None
        """
        self.width = width
        self.bins = bins
        # featuresAll = Analyse_deformed.meanfeaturesC(NPCs, var, circle_allrings)
        # featuresElAll = Analyse_deformed.meanfeaturesE(NPCs, var, ellipse_allrings)
        # _, _, _, _, featuresel3DAll = exportCSV.col_features(NPCs, circle_CRNR, ellipse_CRNR)

        self.rc1 = featuresAll[:, 0]  # circle radius
        self.qe1 = featuresElAll[:, 2]  # minor/major axis ratio
        self.RSSe1 = featuresElAll[:, 4]  # RSS 3D
        self.tilte1 = Analyse_deformed.angles(featuresel3DAll[:, 6:])  # tilt angles
        self.shifte1, self.diste1 = Analyse_deformed.findshiftEl(featuresel3DAll)
        self.featureHist(
            [self.rc1, self.qe1, self.RSSe1, self.tilte1, self.shifte1, self.diste1]
        )

    #   return [rc1, qe1, RSSe1, tilte1, shifte1, diste1]

    def featureHist(self, features1):
        rows = 3
        cols = 2
        name = [
            "radius [nm]",
            "minor/major ellip.",
            "RSS ellip. [nm]",
            "tilt-dif. ellip. [rad]",
            "shift ellip. [nm]",
            "dist ellip [nm]",
        ]

        plt.rcParams.update({"font.size": self.width * 2})

        fig, ax = plt.subplots(
            rows, cols, figsize=(self.width * 0.8, self.width), sharey=False
        )
        # bins = int(len(features1[0])/10) #50

        data = features1[0]

        if self.bins:
            bins = self.bins
        elif data.max() - data.min():
            # determine number of bins via Freedman-Diaconis

            iqr = np.subtract(*np.percentile(data, [75, 25]))
            binwidth = 2 * iqr * (len(data) ** (-1 / 3))
            bins = int((data.max() - data.min()) / binwidth)
            # print(bins)
        else:
            bins = 20

        kwargs = dict(histtype="stepfilled", alpha=0.5, bins=bins, ec="k", linewidth=4)

        for i in range(rows):
            for j in range(cols):
                k = int(i * 2 + j * 1)

                if k == 0:  # set label only for first plot
                    label1 = "$D_{mag}$: 1, $r_\sigma$: 2"
                else:
                    label1 = None

                ax[i, j].hist(features1[k], color="teal", label=label1, **kwargs)

                # ax[i,j].set_ylim([0, 200])
                # ax[i,j].set_yticks([0, 100])
                ax[i, j].set_xlabel(name[k])

        ax[1, 0].set_ylabel("n NPCs")

        def pdffeats(features, i, j, color, label=None, linestyle="-"):
            # i and j are ax indices
            mu, std = norm.fit(features)

            # find area under curve
            values, bns, _ = ax[i, j].hist(features, bins=bins, alpha=0, linewidth=0)
            area = sum(np.diff(bns) * values)

            xmin, xmax = min(bns), max(bns)
            x = np.linspace(xmin, xmax, 100)
            p = norm.pdf(x, mu, std) * area  # (xmax-xmin)/bins
            # label = None
            ax[i, j].plot(x, p, color, linewidth=5, linestyle=linestyle, label=label)

        color1 = "darkcyan"
        pdffeats(features1[0], 0, 0, color1)
        pdffeats(features1[1], 0, 1, color1)
        pdffeats(features1[5], 2, 1, color1)

        fig.legend(loc="upper right")
        # fig.legend.get_frame().set_facecolor((0, 0, 1, 0.1))
        # handles, labels = ax.get_legend_handles_labels()
        # fig.legend(handles, labels, loc='upper center')

        plt.tight_layout()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov  1 18:16:01 2022

@author: maria
"""
import sys
import os
import numpy as np
from gemmi import cif
import pandas as pd
from collections import Counter

# set working directory of input-files and where output files are to be stored
working_dir = "/CoordinatesfromPDB"
working = os.environ.get("WORKING_DIRECTORY", working_dir)
if len(sys.argv) > 1:
    working = sys.argv[1]
os.chdir(working)

# Document of interest
doc = cif.read("7r5k-assembly1.cif")
outputName = "7r5ktest.txt"
output = False
# in the order as they appear in the document (see variable "nups" for reference)
nupnames = [
    "RANBP2",
    "NUP210",
    "ALADIN",
    "NUP93",
    "NUP188",
    "NUP205",
    "NUP155",
    "NDL1",
    "NUP35",
    "NUP54",
    "NUP58NUP45",
    "NUP62",
    "NUP133",
    "NUP107",
    "NUP98NUP96v1",
    "SEC13",
    "SEH1",
    "NUP85",
    "NUP43",
    "NUP160",
    "NUP37",
    "ELYS",
    "NUP98NUP96v2",
    "NUP214",
    "NUP88",
]


block = doc.sole_block()

# Extract relevant data from document
atomID = list(
    block.find_loop("_atom_site.auth_atom_id")
)  # type of atom, we'll later filter for C alpha
seqID = list(
    block.find_loop("_atom_site.auth_seq_id")
)  # index of amino acid of interest
asymID = list(block.find_loop("_atom_site.auth_asym_id"))  # ID of Nup
col_x = list(block.find_values("_atom_site.Cartn_x"))  # Cartesian coordinates of atoms
col_y = list(block.find_values("_atom_site.Cartn_y"))
col_z = list(block.find_values("_atom_site.Cartn_z"))

allvalues = pd.DataFrame([atomID, seqID, asymID, col_x, col_y, col_z]).transpose()
allvalues.columns = ["atomID", "seqID", "asymID", "x", "y", "z"]

allvalues = allvalues[allvalues.atomID == "CA"]  # Filter for only C alpha atoms

# "-" only to be found in Nups IDs of asymmetric unit 2-8, we're interested in asymmetric unit 1
allvalues = allvalues[allvalues.asymID.str.contains("-") == False]

# allvalues = allvalues[allvalues.asymID.str.contains("-5") == True] #opposite rot unit, only to calculate centre

unique_id = np.unique(allvalues.asymID)  # Unique array list of Nup IDs of interest

# Extract indices of N and C terminal resolved amino acids of of C alpha atoms of Nups of interest
indicesC = []
indicesN = []

for value in unique_id:
    indicesN.append(
        allvalues.asymID[allvalues.asymID == value].index[0]
    )  # first aa [0] is most N-terminal
for value in unique_id:
    indicesC.append(
        allvalues.asymID[allvalues.asymID == value].index[-1]
    )  # last aa [-1] is most C-terminal

valuesN = allvalues.filter(items=indicesN, axis=0)
valuesC = allvalues.filter(items=indicesC, axis=0)

nups = np.unique([str(nup[0]) for nup in list(unique_id)])  # nup IDs (types only)
nupsall = np.array(
    [str(nup[0]) for nup in list(unique_id)]
)  # nup IDs, repeated by their frequency

nupsdict = {
    nups[i]: nupnames[i] for i in range(len(nups))
}  # dictionary, keys are nups IDs, entries indicate their names
count = Counter(
    nupsall
)  # dictionary, keys are nup IDs, entries indicate their frequency

if (
    output
):  # print Nup coordinates into a .txt file in a format that can be copy/pasted to Nups_Info.py
    with open(outputName, "w") as f:
        k = 1  # count for nup of a specific type
        ref = ""
        mode = "N"
        idlist = []

        def writeauths(i, values):
            auth = values[values.asymID == i][["asymID"]]
            auth = auth.to_string(index=False, header=False)

            entry = values[values.asymID == i][["x", "y", "z"]]
            entry = entry.to_csv(index=False, header=False).strip()
            return auth, entry

        for i in unique_id:  # for each type of Nup
            if k == 1:  # before a new type of Nup, any terminus (C or N)
                f.write('if (nup == "' + nupsdict[i[0]] + '"):\n')

            #### N-terminus
            if mode == "N":
                idlist.append(i)  # generate list of IDs to later repeat for C-terminus
                auth, entry = writeauths(i, valuesN)
                authnameN = "authN_" + auth

                if k == 1:  # before first Nup of a type, N-terminus
                    f.write('\tif (term == "N"):\n')

                f.write(
                    "\t\t" + authnameN + " = np.array([" + entry + "])\n"
                )  # coordinates

                ref = ref + authnameN + ", "

                # after last Nup of a type, N-terminus
                if k == count[i[0]]:  # counts number of Nups of a specific type
                    f.write(
                        "\t\tref = np.array([" + ref[:-2] + "])\n\n"
                    )  # -2 to remove trailing comma and whitespace
                    k = 0
                    ref = ""
                    mode = "C"  # switch to C terminus once done with N-terminus
                k += 1

            #### C-terminus
            if mode == "C":
                for (
                    i
                ) in (
                    idlist
                ):  # repeat what's been written for N-terminus with C-terminus
                    auth, entry = writeauths(i, valuesC)
                    authnameC = "authC_" + auth

                    if k == 1:  # before first Nup of a type, C-terminus
                        f.write('\tif (term == "C"):\n')

                    f.write(
                        "\t\t" + authnameC + " = np.array([" + entry + "])\n"
                    )  # coordinates

                    ref = ref + authnameC + ", "

                    # after last Nup of a type, C-terminus
                    if k == count[i[0]]:
                        f.write("\t\tref = np.array([" + ref[:-2] + "])\n\n")
                        k = 0
                        ref = ""
                        mode = "N"  # switch back to C-terminus
                        idlist = []
                    k += 1

            # f.write("\n")

# Cstart = {}
# Nstart = {}
# for i in nups:

#     pattern= "^"+str(i)

#     Cstart[pattern] = valuesC[valuesC.asymID.str.contains(pat=pattern, regex=True)].seqID
#     Nstart[pattern] = valuesN[valuesN.asymID.str.contains(pat=pattern,regex=True)].seqID



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar  7 16:21:00 2023

@author: maria
"""

from datetime import date
import uuid
import csv
import math
import Analyse_deformed
import numpy as np
from itertools import compress


def names(var):
    """
    Generates a name dictionary and a unique name for a simulation experiment described by var. 
    The name dictionary contains changed parameters to be written into metadata file. 
    :param var: Dictionary of simulation parameters
    :returns: nameDict: Dictionary of simulation features. Default features are saved as empty strings. 
    name: Unique part of name of a given simulation experiment 
    """
    def makeNames(mean, sigma, meanName, sigmaName, forTitle=True):
        meanStr = str(meanName) + "_" + str(mean) + "_" if mean else ""
        sigmaStr = str(sigmaName) + "_" + str(sigma) + "_" if sigma else ""
        return meanStr + sigmaStr

    today = date.today()
    rStr = makeNames(var["rnew"], var["rsigma"], "r_mean", "r_sigma")
    dStr = makeNames(var["dnew"], var["dsigma"], "dist_mean", "dist_sigma")
    thetaStr = makeNames(
        var["thetanew"], var["thetasigma"], "twist_ang_mean", "twist_ang_sigma"
    )
    elongStr = makeNames(
        var["elliptnew"], var["elliptsigma"], "elong_mean", "elong_sigma"
    )
    symmetStr = "_symmet_" + str(var["symmet"]) + "_" if var["symmet"] != 8 else ""
    kappaStr = makeNames(var["kappa"], None, "kappa", None)
    shiftStr = makeNames(var["shiftsigma"], None, "ring-shift_sigma", None)

    nup_index = ["Ch_" + str(i) for i in list(range(len(var["term"])))]
    thetaref = (
        "theta_ref:_" + var["nup"][0] + "_" + var["term"][0] + "_" if var["rel"] else ""
    )

    channels = [
        sub[item]
        for item in range(len(var["term"]))
        for sub in [nup_index, list(var["nup"]), list(var["term"])]
    ]
    channels = ["_".join(channels)][0]
    modelname = "model_" + str(var["model"])

    # name = (channels+ "_x" + str(n) + modelname
    #     + "_deformmag_" + str(mag) + "_" +  symmetStr + rStr + dStr + thetaStr + thetaref +
    #     elongStr + kappaStr + shiftStr + "seed_" + str(seed) + "_" + today.strftime("%d-%b-%Y"))

    uuid_name = str(uuid.uuid1())
    name = (
        channels
        + "_x"
        + str(var["n"])
        + "_"
        + modelname
        + "_deformmagxy_"
        + str(var["mag"])
        + "_"
        + uuid_name
        + "_"
        + today.strftime("%d-%b-%Y")
    )

    return {
        "rStr": rStr,
        "dStr": dStr,
        "thetaStr": thetaStr,
        "thetaref": thetaref,
        "elongStr": elongStr,
        "symmetStr": symmetStr,
        "kappaStr": kappaStr,
        "shiftStr": shiftStr,
        "channels": channels,
        "modelname": modelname,
    }, name


class MakeCSV:
    def __init__(self, var, NPCs, printNPC, nameDict, name, data_dir):
        """Export a CSV file and a file containing metadata 
       :param var: Dictionary of simulation parameters
       :param NPCs: Dictionary containing simulated NPCs and their metadata
       :param printNPC: NPC coordinates to be exported, so either offsetNPCs or NPCscoords. 
       offsetNPCs should be chosen if the downstream photophysics simulation software does not introduce an offset to structures, 
       NPCscoords should be chosen if downstream photophysics simulation software introduces an offset to structures. 
       :param nameDict: Dictionary of changed parameters 
       :param name: Unique part of name to save output csv file and metadata file 
       :param data_dir: Directory for output files
        
       """

        self.csvpath = data_dir + name + ".csv"
        with open(self.csvpath, "w", newline="") as csvfile:
            writer1 = csv.writer(
                csvfile, delimiter=",", quotechar="|", quoting=csv.QUOTE_MINIMAL
            )
            writer1.writerow(["x", "y", "z", "channel", "particle"])

            for i in range(len(printNPC)):
                # writer1.writerow([printNPC[i,0], printNPC[i,1], printNPC[i,2], printNPC[i,3].astype(int)])
                writer1.writerow(
                    list(printNPC[i, :3])
                    + [printNPC[i, 3].astype(int)]
                    + [printNPC[i, 4].astype(int)]
                )

        # corresponding metadata file
        f = open(data_dir + name + "_metadata.txt", "x")

        # f.write(str(term) + "_" + str(nup) + "_" + str(model) + "\n")
        f.write(nameDict["channels"] + "\n")
        f.write(nameDict["modelname"] + "\n")

        f.write("random seed: " + str(var["seed"]) + "\n")
        f.write("\nmodel layout: ###########\n")

        # f.write("kr = 0.7 \n")
        f.write("nConnect: " + str(var["nConnect"]) + "\n")

        f.write("\nNPC parameters: ##################")

        f.write("\nIrregular deformation parameters: ###########\n")

        f.write("deformation magnitude, xy: " + str(var["mag"]) + "\n")
        f.write("deformation magnitude, z: " + str(var["zmag"]) + "\n")
        f.write("sigma multiplier " + str(var["sigmamult"]) + "\n")

        f.write("\nGeometric NPC properties (input values): ###########\n")
        f.write("symmetry: " + str(var["symmet"]) + "\n")
        f.write(
            "Nup index per ring: (corresponding to channel): "
            + str(NPCs["nupIndex"][0])
            + "\n"
        )
        f.write(
            "Ring membership to subcomplex: " + str(list(NPCs["ringmember"])) + "\n"
        )
        f.write("expected radii per ring: " + str(NPCs["rexp"]) + "\n")
        f.write("expected z per ring: " + str(NPCs["zexp"]) + "\n")
        f.write(
            "expected ring angles per ring (rad): " + str(NPCs["ringAnglesExp"]) + "\n"
        )
        f.write("twist angle input (rad): " + str(NPCs["thetaold"]) + "\n")
        f.write(
            "Nearest rotational unit, NR - CR (degrees): "
            + str(math.degrees(NPCs["theta_offset"]))
            + "\n"
        )
        f.write("\n")

        if nameDict["rStr"]:
            f.write("changed radii: " + nameDict["rStr"].replace("_", " ") + "\n")
        if nameDict["dStr"]:
            f.write("changed distance: " + nameDict["dStr"].replace("_", " ") + "\n")
        if nameDict["elongStr"]:
            f.write(
                "minor/major axis input: "
                + nameDict["elongStr"].replace("_", " ")
                + "\n"
            )
        if nameDict["thetaref"]:
            f.write(
                "ref nup for ring-angles and rotation, "
                + nameDict["thetaref"].replace("_", " ")
                + "\n"
            )
        if nameDict["thetaStr"]:
            f.write(
                "twist angles (rad): " + nameDict["thetaStr"].replace("_", " ") + "\n"
            )
        if nameDict["kappaStr"]:
            f.write(
                "kappa of von mises-fisher distribution: " + str(var["kappa"]) + "\n"
            )
        if nameDict["shiftStr"]:
            f.write(
                "shift of rings (sigma of 0-centered normal, nm): "
                + str(var["shiftsigma"])
                + "\n"
            )
        f.close()


class featuresCSV:
    def __init__(self, NPCs, var, name, circle_allrings, ellipse_allrings, data_dir):
        """
        Generate a CSV file that contains averaged features per NPC. 
        :param var: Dictionary of simulation parameters
        :param NPCs: Dictionary containing simulated NPCs and their metadata
        :param name: Unique part of name to save output csv file and metadata file 
        :param circle_allrings: Features of circles fitted to all NPC rings, these will be averaged per NPC. 
        :param ellipse_allrings: Features of ellipses fitted to all NPC rings, these will be averaged per NPC. 
        :param data_dir: Directory for output files
        """
        
        c_name = [
            "c_r",
            "c_sse",
            "c_sselat",
            "c_sseax",
            "c_x0",
            "c_y0",
            "c_z0",
            "c_tilt_x",
            "c_tilt_y",
            "c_tilt_z",
        ]
        el_name = [
            "el_major",
            "el_minor",
            "el_q",
            "el_rot",
            "el_ssum",
            "el_ssumXY",
            "el_ssumZ",
        ]

        self.csvpath = data_dir + "NPC_features_" + name + ".csv"
        with open(self.csvpath, "w", newline="") as csvfile:
            writer2 = csv.writer(
                csvfile, delimiter=",", quotechar="|", quoting=csv.QUOTE_MINIMAL
            )

            writer2.writerow(
                ["NPC"]
                + c_name
                + el_name
                + ["ce_x0", "ce_y0", "cez0"]
                + ["e_tilt_x", "e_tilt_y", "e_tilt_z"]
            )

            featuresAll, centreAll, tiltAll = Analyse_deformed.meanfeaturesC(
                NPCs, var, circle_allrings, full = True
            )
            featuresElAll, centreElAll, tiltElAll = Analyse_deformed.meanfeaturesE(
                NPCs, var, ellipse_allrings, el_name, full = True
            )

            for i in range(var["n"]):
                writer2.writerow(
                    [i]
                    + list(featuresAll[i])
                    + list(centreAll[i])
                    + list(tiltAll[i])
                    + list(featuresElAll[i])
                    + list(centreElAll[i])
                    + list(tiltElAll[i])
                )


def col_features(NPCs, circle_CRNR, ellipse_CRNR):
    """Compute features per subcomplex
    :param NPCs: Dictionary containing simulated NPCs and their metadata
    :param circle_CRNR: features of circles fitted to subcomplexes 
    :param ellipse_CRNR: featurs of ellipses fitted to subcomplexes 
    :returns: entriesAll: radius,  mean sum of squared errors (SSE), SSE in lateral direction, 
    SSE in axial direction of circles fitted to subcomplexes, columns grouped by type of feature 
    entriescenAll: (n NPCs, (3 dim * n subcomplexes)) centre coordinates per subcomplex, determined by fitting circles 
    entriestiltAll: (n NPCs, (3 dim * n subcomplexes)) tilt coordinates per subcomplex, determined by fitting circles 
    featureselAll: of fitted ellipse per subcomplex: major axis, minor axis, ratio minor/major axis, azimuthal rotation angle, 
    mean sum of squared errors (SSE), SSE in lateral direction, SSE in axial direction, columns grouped by feature 
    """
    _, idx = np.unique(NPCs["ringmember"], return_index=True)
    dim = 3  # dimensions
    memberof = list(
        NPCs["ringmember"][np.sort(idx)]
    )  # list of subcomplexes, e.g. ['CR', 'NR']
    subs = len(memberof)
    n = len(NPCs["NPCs"])
    
    
    featEl = [
        "el_minor",
        "el_major",
        "el_q",
        "el_rot",
        "el_ssum",
        "el_ssumXY",
        "el_ssumZ",
    ]
    
    nfEl = len(featEl)

    feat1D = ["r", "SSE", "SSE_l", "SSE_a"]  # 1D features circle
    entriesAll = np.zeros((n, len(feat1D) * subs))
    entriescenAll = np.zeros((n, dim * subs))
    entriestiltAll = np.zeros((n, dim * subs))
    featureselAll = np.zeros((n, (nfEl) * subs))

    for npc in range(n):
        entriesAll[npc] = np.array(
            [circle_CRNR[sub][npc][i] for i in range(len(feat1D)) for sub in memberof]
        )  # e.g. CR, NR, CR, NR ...
        entriescenAll[npc] = np.array(
            [
                circle_CRNR[sub][npc][4][i]
                for sub in memberof
                for i in range(len(circle_CRNR[sub][npc][4]))
            ]
        )
        entriestiltAll[npc] = np.array(
            [
                circle_CRNR[sub][npc][5][i]
                for sub in memberof
                for i in range(len(circle_CRNR[sub][npc][5]))
            ]
        )
        
        entriesel = [
            ellipse_CRNR[sub][npc][featEl[i]] for i in range(nfEl) for sub in memberof
        ]  # entries alterate between subcomplexes
        
        featureselAll[npc] = np.array(entriesel)  # 'el_minor', 'el_major', 'el_q', 'el_rot', 'el_ssum', 'el_ssumXY', 'el_ssumZ'
        

    featuresel3DAll = colfeaturesEl3D(NPCs, ellipse_CRNR)
    
    return entriesAll, entriescenAll, entriestiltAll, featureselAll, featuresel3DAll



def colfeaturesEl3D(NPCs, ellipse_CRNR):
    """ Tilt and centre coordinates of subcomplexes, determined by fitting ellipses 
    :param NPCs: Dictionary containing simulated NPCs and their metadata
    :param circle_CRNR: features of circles fitted to subcomplexes
    :returns: featuresel3DAll: centre coordinates per subcomplex, tilt coordinates per subcomplex, determined by fitting circles.  
    """
    _, idx = np.unique(NPCs["ringmember"], return_index=True)
    memberof = list(
        NPCs["ringmember"][np.sort(idx)]
    )  # list of subcomplexes, e.g. ['CR', 'NR']
    dim = 3  # dimensions

    featEl = ["Ce", "normal2",]
    nfEl = len(featEl)
    subs = len(memberof)
    n = len(NPCs["NPCs"])
    featuresel3DAll = np.zeros((n, 2 * dim * subs))


    for npc in range(n):     
        entriesel = [
            ellipse_CRNR[sub][npc][featEl[i]] for i in range(nfEl) for sub in memberof
        ]  # entries alterate between subcomplexes
        
        featuresel3DAll[npc] = np.array(entriesel).flatten()  # 'Ce', 'normal2', alternating by subcomplex
        
    return featuresel3DAll


class featuresCSV_subcomplex:
    def __init__(self, NPCs, circle_CRNR, ellipse_CRNR, name, data_dir):
        _, idx = np.unique(NPCs["ringmember"], return_index=True)
        memberof = list(NPCs["ringmember"][np.sort(idx)])
        """
        Generate a CSV file that contains features for each NPC subcomplex. 
        :param NPCs: Dictionary containing simulated NPCs and their metadata
        :param circle_CRNR: Features of circles fitted to the NPC subcomplexes. 
        :param ellipse_CRNR: Features of ellipses fitted to the NPC subcomplexes. 
        :param name: Unique part of name to save output csv file and metadata file 
        :param data_dir: Directory for output files
        """

        col = ["r", "SSE", "SSE_l", "SSE_a"]
        col2 = ["Cx", "Cy", "Cz"]
        col3 = ["Tx", "Ty", "Tz"]

        columns = [c + "_" + mem for c in col for mem in memberof]
        cen = [c + "_" + mem for mem in memberof for c in col2]
        tilt = [c + "_" + mem for mem in memberof for c in col3]

        colel = [
            "el_minor",
            "el_major",
            "el_q",
            "el_rot",
            "el_ssum",
            "el_ssumXY",
            "el_ssumZ",
            "Ce",
            "normal2",
        ]
        columnsel = [c + "_" + mem for c in colel[:7] for mem in memberof]
        cenel = ["el_" + c + "_" + mem for mem in memberof for c in col2]
        tiltel = ["el_" + c + "_" + mem for mem in memberof for c in col3]

        self.csvpath = data_dir + "NPC_features_subcomplex_" + name + ".csv"
        with open(self.csvpath, "w", newline="") as csvfile:  # TODO: change directory
            writer2 = csv.writer(
                csvfile, delimiter=",", quotechar="|", quoting=csv.QUOTE_MINIMAL
            )

            writer2.writerow(
                ["NPC"] + columns + cen + tilt + columnsel + cenel + tiltel
            )

            (
                entriesAll,
                entriescenAll,
                entriestiltAll,
                featureselAll,
                featuresel3DAll,
            ) = col_features(NPCs, circle_CRNR, ellipse_CRNR)

            n = len(NPCs["NPCs"])

            for NPC in range(n):
                writer2.writerow(
                    [NPC]
                    + list(entriesAll[NPC])
                    + list(entriescenAll[NPC])
                    + list(entriestiltAll[NPC])
                    + list(featureselAll[NPC])
                    + list(featuresel3DAll[NPC])
                )


class featuresCSV_rings:
    def __init__(self, NPCs, var, name, data_dir, circle_allrings, ellipse_allrings):
        """
        Generate a CSV file that contains features for each NPC ring. 
        :param NPCs: Dictionary containing simulated NPCs and their metadata
        :param var: Dictionary of simulation parameters
        :param name: Unique part of name to save output csv file and metadata file 
        :param data_dir: Directory for output files
        :param circle_allrings: Features of circles fitted to all NPC rings.
        :param ellipse_allrings: Features of ellipses fitted to all NPC rings. 
        """
        flattenvalues = lambda NPC_n, key, outputrange: np.array(
            [NPC_n[j][key] for j in outputrange]
        ).flatten()

        nfeat = 6  # 6 because: r, sse, sse_l, sse_a, centre, tilt
        nfeatel = 9
        col = ["r", "SSE", "SSE_l", "SSE_a"]
        col2 = ["Cx", "Cy", "Cz"]
        col3 = ["Tx", "Ty", "Tz"]

        NP = [i.replace("up", "") for i in var["nup"]]  # shorten nup names
        z = NPCs["zexp"]
        nupIndex = NPCs["nupIndex"]
        colnames = [
            cl + "_" + NP[nupIndex[0][j]] + "_" + str(round(z[j])) + "nm"
            for cl in col
            for j in range(len(z))
        ]

        cen = [
            cl + "_" + NP[nupIndex[0][j]] + "_" + str(round(z[j])) + "nm"
            for j in range(len(z))
            for cl in col2
        ]
        tilt = [
            cl + "_" + NP[nupIndex[0][j]] + "_" + str(round(z[j])) + "nm"
            for j in range(len(z))
            for cl in col3
        ]
        colnames2 = cen + tilt

        col_e = [
            "el_minor",
            "el_major",
            "el_q",
            "el_rot",
            "el_ssum",
            "el_ssumXY",
            "el_ssumZ",
            "Ce",
            "normal2",
        ]

        colel = [
            "el_minor",
            "el_major",
            "el_q",
            "el_rot",
            "el_ssum",
            "el_ssumXY",
            "el_ssumZ",
        ]
        colel2 = ["eCx", "eCy", "eCz"]
        colel3 = ["eTx", "eTy", "eTz"]

        z_i_ring = np.repeat([i for i in range(var["n"])], len(NPCs["zexp"]))

        colnamesel = [
            cl + "_" + NP[nupIndex[0][j]] + "_" + str(round(z[j])) + "nm"
            for cl in colel
            for j in range(len(z))
        ]
        cenel = [
            cl + "_" + NP[nupIndex[0][j]] + "_" + str(round(z[j])) + "nm"
            for j in range(len(z))
            for cl in colel2
        ]
        tiltel = [
            cl + "_" + NP[nupIndex[0][j]] + "_" + str(round(z[j])) + "nm"
            for j in range(len(z))
            for cl in colel3
        ]
        colnamesel2 = cenel + tiltel

        self.csvpath = data_dir + "NPC_features_rings_" + name + ".csv"
        with open(self.csvpath, "w", newline="") as csvfile:  # TODO: change directory
            writer2 = csv.writer(
                csvfile, delimiter=",", quotechar="|", quoting=csv.QUOTE_MINIMAL
            )

            writer2.writerow(["NPC"] + colnames + colnames2 + colnamesel + colnamesel2)

            for i in range(var["n"]):
                circle_NPC_n = list(compress(circle_allrings, z_i_ring == i))
                ellipse_NPC_n = list(compress(ellipse_allrings, z_i_ring == i))

                range_n_rings_in_i = range(len(circle_NPC_n))

                circleRings = []
                ellipseRings = []
                for k in range(nfeat):
                    circleRings.append(
                        flattenvalues(circle_NPC_n, k, range_n_rings_in_i)
                    )

                for k in range(nfeatel):
                    ellipseRings.append(
                        flattenvalues(ellipse_NPC_n, col_e[k], range_n_rings_in_i)
                    )

                features = np.array(
                    circleRings[: nfeat - 2]
                ).flatten()  # -2 because last two entries are 3D vectors (tilt and centre)
                features3D = np.array(circleRings[nfeat - 2 :]).flatten()

                featuresel = np.array(ellipseRings[: nfeatel - 2]).flatten()
                featuresel3D = np.array(ellipseRings[nfeatel - 2 :]).flatten()

                writer2.writerow(
                    [i]
                    + list(features)
                    + list(features3D)
                    + list(featuresel)
                    + list(featuresel3D)
                )






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 17 18:13:18 2021

@author: maria
"""
import math
import numpy as np
import circle_fit as cf
from warnings import warn
from itertools import compress


def ellipses(NPCscoords, membership=None):
    """Fit ellipses to the NPC or its subcomponents 
    
    :param NPCscoords: Coordinates of the NPC 
    :param membership: information to which part of the NPC ellipses should be fitted. 
    if membership = NPCs["ringmemall"], ellipses will be fitted to subcomplexes. 
    if mempership = NPCs["z_i_all"], ellipses will be fitted to rings. 
    if membership = None, ellipses will be fitted to the whole NPC. Defaults to None. 
    :type membership: NoneType, or numpy.ndarray
    :return: features of fitted ellipses per NPC 
    :rtype: list for membership = NPCs["z_i_all"] or membership = None, dict (for membership = NPCs["ringmemall"])
    
    """
    n = len(np.unique(NPCscoords[:, -1]))  # number of NPCs
    iplus1 = int(NPCscoords[-1][-1] + 1)  # index+1 of last or only NPC

    if n != iplus1 and n != 1:
        warn("Indexing of NPCs off.")
        return

    def squaresum(NPC_i, x0, y0, ellipsefeatures_i, z=None, ring=None):
        _, dist = align2ellipse(NPC_i, x0, y0, ellipsefeatures_i, z, ring)
        return sum([j**2 for j in dist])

    def roundfeatures(ellipsefeatures, z=None, ring=None):
        if z != None:
            suffix = "_z" + str(z)
        elif ring != None:
            suffix = "_" + str(ring)
        else:
            suffix = ""

        ellipsefeatures["el_major" + suffix] = round(
            ellipsefeatures["el_major" + suffix], ndigits=1
        )
        ellipsefeatures["el_minor" + suffix] = round(
            ellipsefeatures["el_minor" + suffix], ndigits=1
        )
        ellipsefeatures["el_ssum" + suffix] = round(
            ellipsefeatures["el_ssum" + suffix], ndigits=1
        )
        ellipsefeatures["el_ssumXY" + suffix] = round(
            ellipsefeatures["el_ssumXY" + suffix], ndigits=1
        )
        ellipsefeatures["el_ssumZ" + suffix] = round(
            ellipsefeatures["el_ssumZ" + suffix], ndigits=1
        )
        ellipsefeatures["el_q" + suffix] = round(
            ellipsefeatures["el_q" + suffix], ndigits=2
        )
        ellipsefeatures["el_rot" + suffix] = round(
            ellipsefeatures["el_rot" + suffix], ndigits=3
        )
        return ellipsefeatures

    if str(membership) == "None":
        ellipsefeatures = [None] * n
        squaresumsXY = [None] * n
        squaresumsZ = [None] * n
        squaresums = [None] * n

        i0 = 0
        for i in range(iplus1 - n, iplus1):
            NPC_i = NPCscoords[NPCscoords[:, 4] == i]
            NPC_xy, x0, y0, ellipsefeatures[i0] = fitEllipse3D(NPC_i)
            squaresumsXY[i0] = squaresum(NPC_xy, x0, y0, ellipsefeatures[i0])
            squaresumsZ[i0] = sum(NPC_xy[:, 2] ** 2)
            squaresums[i0] = squaresumsXY[i0] + squaresumsZ[i0]

            ellipsefeatures[i0].update(
                {
                    "el_ssum": squaresums[i0],
                    "el_ssumXY": squaresumsXY[i0],
                    "el_ssumZ": squaresumsZ[i0],
                }
            )

            ellipsefeatures[i0] = roundfeatures(ellipsefeatures[i0])
            i0 += 1

        return ellipsefeatures

    # Fit ellipse to NPC subcomplexes
    elif type(membership[0]) == np.str_:

        def ringmem(name):
            part = NPCscoords[membership == name]
            elPart = [None] * n
            squaresumsXY = [None] * n
            squaresumsZ = [None] * n
            squaresums = [None] * n

            i0 = 0
            for i in range(iplus1 - n, iplus1):
                NPC_part_i = part[part[:, 4] == i]
                NPC_xy, x0, y0, elPart[i0] = fitEllipse3D(NPC_part_i)

                squaresumsXY[i0] = squaresum(NPC_xy, x0, y0, elPart[i0])
                squaresumsZ[i0] = sum(NPC_xy[:, 2] ** 2)
                squaresums[i0] = squaresumsXY[i0] + squaresumsZ[i0]

                elPart[i0].update(
                    {
                        "el_ssum": squaresums[i0],
                        "el_ssumXY": squaresumsXY[i0],
                        "el_ssumZ": squaresumsZ[i0],
                    }
                )

                elPart[i0] = roundfeatures(elPart[i0])

                i0 += 1
            return elPart

        partDictEl = makePartDict(membership, ringmem)

        return partDictEl  # [ellipseCR, ellipseNR]

    elif type(membership[0]) == np.int64 or type(membership[0] == np.float64):
        ellipseAll = [None] * (membership[-1] - membership[0] + 1)

        squaresumsXY = [None] * (membership[-1] - membership[0] + 1)
        squaresumsZ = [None] * (membership[-1] - membership[0] + 1)
        squaresums = [None] * (membership[-1] - membership[0] + 1)
        i0 = 0

        for i in range(membership[0], membership[-1] + 1):
            NPC_xy, x0, y0, ellipseAll[i0] = fitEllipse3D(NPCscoords[membership == i])
            squaresumsXY[i0] = squaresum(NPC_xy, x0, y0, ellipseAll[i0])
            squaresumsZ[i0] = sum(NPC_xy[:, 2] ** 2)
            squaresums[i0] = squaresumsXY[i0] + squaresumsZ[i0]

            ellipseAll[i0].update(
                {
                    "el_ssum": squaresums[i0],
                    "el_ssumXY": squaresumsXY[i0],
                    "el_ssumZ": squaresumsZ[i0],
                }
            )

            ellipseAll[i0] = roundfeatures(ellipseAll[i0])
            i0 += 1

        return ellipseAll


def fitEllipse3D(NPC, z=None, ring=None):
    NPC_centered, V, Ce = fitPlane(
        NPC[:, :3]
    )  # find plane that minimises distance to points

    normal0 = V[0, :]
    normal1 = V[1, :]
    normal2 = V[2, :]

    basisZ = np.array([0, 0, 1])

    # planar rotation only necessary if SVD detects varation in Z
    rotate = False if np.array_equal(basisZ, abs(normal2)) else True
    NPC_xy = rodrigues_rot(NPC_centered, normal2, basisZ) if rotate else NPC_centered  #

    ### fit ellipse parameters to planar NPC
    X = np.array([NPC_xy[:, 0]]).T  # X coordinates NPC
    Y = np.array([NPC_xy[:, 1]]).T  # Y

    # Formulate and solve the least squares problem ||A0x - b ||^2
    A0 = np.hstack([X**2, X * Y, Y**2, X, Y])
    b0 = np.ones_like(X)

    x = np.linalg.lstsq(A0, b0, rcond=None)[0].squeeze()

    A, B, C, D, E = x
    F = -1

    a = -np.sqrt(
        2
        * (A * E**2 + C * D**2 - B * D * E + (B**2 - 4 * A * C) * F)
        * ((A + C) - np.sqrt((A - C) ** 2 + B**2))
    ) / (B**2 - 4 * A * C)

    b = -np.sqrt(
        2
        * (A * E**2 + C * D**2 - B * D * E + (B**2 - 4 * A * C) * F)
        * ((A + C) + np.sqrt((A - C) ** 2 + B**2))
    ) / (B**2 - 4 * A * C)

    x0 = (2 * C * D - B * E) / (
        B**2 - 4 * A * C
    )  # centre of ellipse fitted to NPC_xy. Not necessarily 0, 0
    y0 = (2 * A * E - B * D) / (B**2 - 4 * A * C)

    minor = min(a, b)  # length of minor axis
    major = max(a, b)  # length of major axis

    if B != 0:
        rho = np.arctan(1 / B * (C - A - np.sqrt((A - C) ** 2 + B**2)))
    elif A < C:
        rho = 0
    elif A > C:
        rho = 0.5 * np.pi
    if A < 0:
        rho -= 0.5 * np.pi
        if rho < -0.5 * np.pi:
            rho += np.pi

    q = minor / major

    # points where minor and major axes intercept ellipse fitted to NPC_xy
    xmajor = major * np.cos(rho)
    ymajor = major * np.sin(rho)
    xminor = minor * np.cos(rho - np.pi / 2)
    yminor = minor * np.sin(rho - np.pi / 2)

    # tilted same as NPC, but not offset. Offset applied later via Ce

    if rotate:
        majv = rodrigues_rot(np.array([xmajor, ymajor, 0]), basisZ, normal2)
        minv = rodrigues_rot(np.array([xminor, yminor, 0]), basisZ, normal2)
        Ce = Ce + rodrigues_rot(np.array([x0, y0, 0]), basisZ, normal2)

    else:
        majv = np.array([xmajor, ymajor, 0])
        minv = np.array([xminor, yminor, 0])

    Ce = Ce.flatten()

    keyList = [
        "el_minor",
        "el_major",
        "el_q",
        "el_rot",
        "minv",
        "majv",
        "Ce",
        "normal2",
        "normal1",
        "normal0",
    ]
    valueList = [minor, major, q, rho, minv, majv, Ce, normal2, normal1, normal0]

    if z != None:
        keyList = [i + "_z" + str(z) for i in keyList]
    elif ring != None:
        keyList = [i + "_" + str(ring) for i in keyList]

    zip_iterator = zip(keyList, valueList)
    return NPC_xy, x0, y0, dict(zip_iterator)


def dist2ellipse(semi_major, semi_minor, p):
    # https://github.com/0xfaded/ellipse_demo/issues/1
    # project all points onto 1st quadrant
    px = abs(p[0])
    py = abs(p[1])

    t = math.pi / 4

    a = semi_major
    b = semi_minor

    for _ in range(0, 3):  # 3 iterations
        x = a * math.cos(t)
        y = b * math.sin(t)

        ex = (a * a - b * b) * math.cos(t) ** 3 / a
        ey = (b * b - a * a) * math.sin(t) ** 3 / b

        rx = x - ex
        ry = y - ey

        qx = px - ex
        qy = py - ey

        r = math.hypot(ry, rx)
        q = math.hypot(qy, qx)

        delta_c = r * math.asin((rx * qy - ry * qx) / (r * q))
        delta_t = delta_c / math.sqrt(a * a + b * b - x * x - y * y)

        t += delta_t
        t = min(math.pi / 2, max(0, t))

    closestpoint = (
        math.copysign(a * math.cos(t), p[0]),
        math.copysign(b * math.sin(t), p[1]),
    )
    return closestpoint, np.linalg.norm(p - closestpoint)


def rotate(origin, point, angle):
    """
    Rotate a point counterclockwise by a given angle around a given origin.
    The angle should be given in radians.
    """
    ox, oy = origin
    px, py = point

    qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
    qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
    return qx, qy


def align2ellipse(NPC, x0, y0, elfeatures, z=None, ring=None):
    # elfeatures = elfeatures[0]
    #    "x0" : x0, "y0" : y0, "minor" : minor, "major" : major, "q": q, "rho" : rho}

    if z != None:
        suffix = "_z" + str(z)
    elif ring != None:
        suffix = "_" + str(ring)
    else:
        suffix = ""

    minor = elfeatures["el_minor" + suffix]
    major = elfeatures["el_major" + suffix]
    rho = elfeatures["el_rot" + suffix]

    nodes = len(NPC)
    NPCrot = np.array(NPC)

    # rotate around centre of fitted ellipse so that the major axis aligns with
    # the y axis
    for i in range(nodes):
        NPCrot[i][0:2] = rotate([x0, y0], NPC[i][0:2], -rho)

    # offset so that the centre is at 0
    NPCrot[:, 0] -= np.mean(NPCrot[:, 0])
    NPCrot[:, 1] -= np.mean(NPCrot[:, 1])

    dist = [None] * nodes
    closestpoints = [None] * nodes

    for i in range(nodes):
        closestpoints[i], dist[i] = dist2ellipse(major, minor, NPCrot[i][0:2])

    return closestpoints, dist


#########################################################################


def circles(NPCscoords, membership=None):
    """Fit circles to the NPC or its subcomponents 
    
    :param NPCscoords: Coordinates of the NPC 
    :param membership: information to which part of the NPC circles should be fitted. 
    if membership = NPCs["ringmemall"], circles will be fitted to subcomplexes. 
    if mempership = NPCs["z_i_all"], circles will be fitted to rings. 
    if membership = None, circles will be fitted to the whole NPC. Defaults to None. 
    :type membership: NoneType, or numpy.ndarray
    :return: features of fitted circles per NPC 
    :rtype: list for membership = NPCs["z_i_all"] or membership = None, dict (for membership = NPCs["ringmemall"])
    
    """
    n = len(np.unique(NPCscoords[:, -1]))  # number of NPCs
    iplus1 = int(NPCscoords[-1][-1] + 1)  # index+1 of last or only NPC

    if n != iplus1 and n != 1:
        warn("Indexing of NPCs off.")
        return

    # if str(ringmode) == "None":
    if str(membership) == "None":
        circlefeatures = [None] * n

        i0 = 0
        for i in range(iplus1 - n, iplus1):
            circlefeatures[i0] = fitCircle3D(
                NPCscoords[NPCscoords[:, 4] == i]
            )  # TODO: 4
            i0 += 1
        return circlefeatures  # {"circlefeatures": circlefeatures} #TODO: Maybe change back

    # if ringmode == "CRNR":
    elif type(membership[0]) == np.str_:
        #
        def ringmem(name):  # name is "CR", "NR", BRCR", "BRNR" or "IR"
            part = NPCscoords[membership == name]
            circlePart = [None] * n
            i0 = 0
            for i in range(iplus1 - n, iplus1):
                circlePart[i0] = fitCircle3D(part[part[:, 4] == i])
                i0 += 1
            return circlePart

        partDict = makePartDict(membership, ringmem)

        return partDict

    # if ringmode == "z":
    elif type(membership[0]) == np.int64 or type(membership[0] == np.float64):
        circlefeaturesZ = [None] * (membership[-1] - membership[0] + 1)
        i0 = 0

        for i in range(membership[0], membership[-1] + 1):
            circlefeaturesZ[i0] = fitCircle3D(NPCscoords[membership == i])
            i0 += 1

        return circlefeaturesZ  # {"circleAll" : circleAll}


def makePartDict(membership, ringmem):
    partDict = {}
    if "CR" in membership:
        partDict["CR"] = ringmem("CR")

    if "NR" in membership:
        partDict["NR"] = ringmem("NR")

    if "BRCR" in membership:
        partDict["BRCR"] = ringmem("BRCR")

    if "BRNR" in membership:
        partDict["BRNR"] = ringmem("BRNR")

    if "IR" in membership:
        partDict["IR"] = ringmem("IR")

    return partDict


def fitCircle3D(NPC):
    NPC = NPC[:, :3]  ##############

    NPC_centered, V, C = fitPlane(NPC)

    normal0 = V[0, :]
    normal1 = V[1, :]
    normal2 = V[2, :]  # if no variation in z, normal2= [0, 0, 1]

    basisZ = np.array([0, 0, 1])
    rotate = False if np.array_equal(basisZ, normal2) else True

    NPC_xy = rodrigues_rot(NPC_centered, normal2, basisZ) if rotate else NPC_centered  #

    zsqsum = sum(NPC_xy[:, 2] ** 2)  # square sum of distances from circle-plane in z

    xc, yc, r, residual = cf.least_squares_circle(
        NPC_xy[:, :2]
    )  # xcentre, ycentre, radius

    sqsum = zsqsum + residual

    if rotate:
        C = C + rodrigues_rot(
            np.array([xc, yc, 0]), basisZ, normal2
        )  # Transform circle center back to 3D coords

    C = C.flatten()

    return [
        r,
        sqsum,
        residual,
        zsqsum,
        C,
        normal2,
        normal1,
        normal0,
    ]  # [xc, yc, np.round(r, 2), residual] # TODO: change back


def fitPlane(P):
    P_mean = P.mean(axis=0)  # centre
    P_centered = P - P_mean  # 0-centre
    U, s, V = np.linalg.svd(P_centered)
    return P_centered, V, P_mean


def rodrigues_rot(P, n0, n1):
    # If P is only 1d array (coords of single point), fix it to be matrix
    if P.ndim == 1:
        P = P[np.newaxis, :]

    # Get vector of rotation k and angle theta
    n0 = n0 / np.linalg.norm(n0)
    n1 = n1 / np.linalg.norm(n1)

    k = np.cross(n0, n1)
    k = k / np.linalg.norm(k)
    theta = np.arccos(np.dot(n0, n1))

    # Compute rotated points
    P_rot = np.zeros((len(P), 3))
    for i in range(len(P)):
        P_rot[i] = (
            P[i] * np.cos(theta)
            + np.cross(k, P[i]) * np.sin(theta)
            + k * np.dot(k, P[i]) * (1 - np.cos(theta))
        )

    return P_rot


def fitCircle(NPC):
    xc, yc, r, residual = cf.least_squares_circle(
        NPC
    )  # xcentre, ycentre, radius #TODO: cf only works in 2D
    xc = round(xc, ndigits=1)
    yc = round(yc, ndigits=1)
    # r = round(r, ndigits=1) TODO
    residual = round(residual, ndigits=1)
    return [
        xc,
        yc,
        r,
        residual,
    ]  # [xc, yc, np.round(r, 2), residual] # TODO: change back


def meanfeaturesC(NPCs, var, circle_allrings:list, full:bool = False):
    """
    :param NPCs: Dictionary containing simulated NPCs and their metadata
    :param var: Dictionary of simulation parameters
    :param circle_allrings: features of circles fitted to each NPC ring 
    :param full: indicates whether to also return centreAll and tiltAll, default False 
    :return: featuresAll, shape (n NPCs, 4)
    featuresAll[:,0]: mean NPC circle radius 
    featuresAll[:,1]: mean sum of squared errors (SSE) of circles fitted per NPC 
    featuresAll[:, 2]: mean SSE in lateral direction of circles fitted per NPC 
    featuresAll[:, 3]: mean SSE in axial direction of circles fitted per NPC 
    optional: 
    centreAll: shape (n NPCs, 3): mean centre per NPC, determined by fitting circles 
    tiltAll:  shape (n NPCs, 3): mean tilt of NPC, determined by fitting circles 
    """
    z_i_ring = np.repeat(
        [i for i in range(var["n"])], len(NPCs["zexp"])
    )  # index that assigns each ring to an NPC
    n = len(np.unique(z_i_ring))  # n NPCs
    featuresAll = np.zeros((n, 4))  # 4 because radius, SSE, SSE lateral, SSE axial



    for i in range(n):
        circle_NPC_n = list(
            compress(circle_allrings, z_i_ring == i)
        )  # circle_allrings, but only values for NPC with index i
        range_n_rings_in_i = range(
            len(circle_NPC_n)
        )  # range of numbers of rings within NPC i
        featuresAll[i] = np.array(
            [
                np.mean([circle_NPC_n[j][k] for j in range_n_rings_in_i])
                for k in range(4)
            ]
        )  # r, sqsum, residual, zsqsum
        
        
    if full: 
        centreAll = np.zeros((n, 3))  # 3 because 3D
        tiltAll = np.zeros((n, 3))  # 3 because 3D     
        for i in range(n):
                
            
            centreAll[i] = np.mean(
                [circle_NPC_n[j][4] for j in range_n_rings_in_i], axis=0
            )  # centre per NPC
            tiltAll[i] = np.mean(
                [circle_NPC_n[j][5] for j in range_n_rings_in_i], axis=0
            )  # tilt per NPC
        return featuresAll, centreAll, tiltAll
    
    return featuresAll

def meanfeaturesE(NPCs, var, ellipse_allrings:list, el_name=False, full = False):
    """
    :param NPCs: Dictionary containing simulated NPCs and their metadata
    :param var: Dictionary of simulation parameters
    :param ellipse_allrings: features of ellipses fitted to each NPC ring
    :type ellipse_allrings: list of dictionaries
    :param el_name: list of feature names, corrsponding to keys in ellipse_allring. 
    ["el_major", "el_minor", "el_q" "el_rot", "el_ssum", "el_ssumXY", "el_ssumZ"] if False. Defaults to False. 
    :return: 
    featuresElAll: Mean features per NPC determined by averaging ring-wise fitted ellipses. 
        shape (nNPCs, n features), n features is 7 by default, corresponding to el_name: 
        length major axis, length minor axis, ratio minor/major, azimuthal angle ellipse, 
        sum of square errors, sum of square errors in lateral direction, sum of square errors in axial direction
    centreElAll: shape (n NPCs, 3): mean centre per NPC, determined by fitting ellipses 
    tiltElAll:  shape (n NPCs, 3): mean tilt of NPC, determined by fitting ellipses 
    """
    z_i_ring = np.repeat([i for i in range(var["n"])], len(NPCs["zexp"]))

    if el_name == False:
        el_name = [
            "el_major",
            "el_minor",
            "el_q",
            "el_rot",
            "el_ssum",
            "el_ssumXY",
            "el_ssumZ",
        ]
    n = len(np.unique(z_i_ring))  # n NPCs
    featuresElAll = np.zeros((n, len(el_name)))


    for i in range(n):
        ellipse_NPC_n = list(compress(ellipse_allrings, z_i_ring == i))
        range_n_rings_in_i = range(len(ellipse_NPC_n))

        featuresElAll[i] = np.array(
            [
                np.mean([ellipse_NPC_n[j][i] for j in range_n_rings_in_i])
                for i in el_name
            ]
        )
    
    if full: 
        centreElAll = np.zeros((n, 3))  # 3 because 3D
        tiltElAll = np.zeros((n, 3))  # 3 because 3D    
        for i in range(n):
            centreElAll[i] = np.mean(
                [ellipse_NPC_n[i]["Ce"] for i in range_n_rings_in_i], axis=0
            )
            tiltElAll[i] = np.mean(
                [ellipse_NPC_n[i]["normal2"] for i in range_n_rings_in_i], axis=0
            )
    
        return featuresElAll, centreElAll, tiltElAll
    return featuresElAll


def angles(entriestiltAll):
    """Input: n * 6 array of tilt angles, where the first 3 columns are tilt-angles of
    the first sub-complex, and the next 3 colums are tilt-angles of the second sub-complex.
    n is the number of NPCs"""

    def findangle(v1, v2):  # vector 1, vector 2
        uv1 = v1 / np.linalg.norm(v1)
        uv2 = v2 / np.linalg.norm(v2)
        return np.arccos(np.dot(uv1, uv2))

    tiltangles = np.zeros(len(entriestiltAll))

    for i in range(len(entriestiltAll)):
        tiltv1 = entriestiltAll[i][:3]
        tiltv2 = entriestiltAll[i][3:]
        tiltangles[i] = findangle(tiltv1, tiltv2)
        if tiltangles[i] >= np.pi / 2:
            tiltangles[i] = np.pi - tiltangles[i]

    return tiltangles


def findshiftEl(featuresel3DAll):
    c1 = featuresel3DAll[:, :2]
    c2 = featuresel3DAll[:, 3:5]
    d1 = featuresel3DAll[:, 2]
    d2 = featuresel3DAll[:, 5]
    shift = np.linalg.norm(c1 - c2, axis=1)
    dist = d1 - d2
    return shift, dist



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 10 16:43:21 2021

@author: maria
"""

import re
import numpy as np
from warnings import warn

terminus = "N"
rotunit = "multiple"

ChimeraOutput = """
Atom #1.1/G:418@CA 978.520 659.710 677.663
Atom #1.1/M:418@CA 967.247 721.001 720.364
Atom #1.1/S:418@CA 981.755 756.542 750.005
Atom #1.1/Y:418@CA 964.003 697.693 708.018
"""


def CleanCoords(ChimeraOutput, terminus):
    """Used for 5A9Q; rotational unit has to be specified"""

    cleaned = re.sub("#", "", ChimeraOutput)

    # remove whitespaces at start and end of line
    cleaned = re.sub("^ *", "", cleaned, flags=re.MULTILINE)
    cleaned = re.sub(" *$", "", cleaned, flags=re.MULTILINE)

    # remove starting and trailing linebreaks
    while cleaned.index("\n") == 0:
        cleaned = re.sub("^\n", "", cleaned)
        if cleaned.index("\n") != 0:
            cleaned = cleaned[::-1]

    cleaned = re.sub("Atom ", "", cleaned)
    refname = "auth" + terminus + "_"
    cleaned = re.sub("1\../", refname, cleaned)

    cleaned = re.sub(" ", ", ", cleaned)
    cleaned = re.sub(":[0-9]*@CA, ", " = np.array([", cleaned)
    cleaned = re.sub("$", "])", cleaned, flags=re.MULTILINE)

    return cleaned


def CleanCoords2(ChimeraOutput, terminus):
    """Used for 7PEQ; rotational unit does not need to be specified"""
    cleaned = re.sub("", "", ChimeraOutput)

    # remove whitespaces at start and end of line
    cleaned = re.sub("^ *", "", cleaned, flags=re.MULTILINE)
    cleaned = re.sub(" *$", "", cleaned, flags=re.MULTILINE)

    # remove starting and trailing linebreaks
    while cleaned.index("\n") == 0:
        cleaned = re.sub("^\n", "", cleaned)
        if cleaned.index("\n") != 0:
            cleaned = cleaned[::-1]

    cleaned = re.sub("Atom ", "", cleaned)
    refname = "auth" + terminus + "_"
    cleaned = re.sub("/", refname, cleaned)

    cleaned = re.sub(" ", ", ", cleaned)
    cleaned = re.sub(":[0-9]*@CA, ", " = np.array([", cleaned)
    cleaned = re.sub("$", "])", cleaned, flags=re.MULTILINE)

    return cleaned


def summarise_refs(cleaned):
    refslist = re.findall("auth.*=", cleaned)
    refslist = [re.sub("=", "", ref) for ref in refslist]
    refslist = [re.sub(" ", "", ref) for ref in refslist]

    refs = ", ".join([str(elem) for elem in refslist])
    return "ref = np.array([" + refs + "])"


if rotunit == "multiple":
    cleaned = CleanCoords(ChimeraOutput, terminus)
if rotunit == "one":
    cleaned = CleanCoords2(ChimeraOutput, terminus)

refs = summarise_refs(cleaned)
print(cleaned)
print(refs)


def findCentre(nup, nupop):
    """
    nup: Coordinates of a nup of interest
    nupop: Coordinates of the oposite nup of interest.
    For an 8-fold symmetric NPC, the index of an opposite nup is
    + 4.
    e.g. getcrd #1.1/A:20@CA
    and getcrd #1.5/A:20@CA
    """
    # if nup[2] != nupop[2]:
    #     warn("Opposite nups must have the same z coordinates.")
    #     return

    x = np.mean((nup[0], nupop[0]))
    y = np.mean((nup[1], nupop[1]))

    return x, y


nup0 = np.array([1127.765, 714.280, 844.037])
nup1 = np.array([306.035, 712.820, 844.037])
# (716.9000000000001, 713.55)


nup01 = np.array([1035.081, 602.042, 702.602])
nup02 = np.array([398.719, 825.058, 702.602])
# (716.9, 713.55)

nup03 = np.array([828.408, 1031.731, 702.602])
nup07 = np.array([605.392, 395.369, 702.602])
# (716.9000000000001, 713.55)

NupO0 = np.array([1035.081, 602.042, 702.602])
NupO1 = authC_E = np.array([398.719, 825.058, 702.602])

nup7per = np.array([1143.648, 766.967, 1053.077])
nup7per2 = np.array([1045.163, 1421.828, 1052.971])
findCentre(nup7per, nup7per2)


nup107N_00_7r5k = np.array([559.908, 768.369, 1246.970])
nup107N_05_7r5k = np.array([1381.212, 1172.751, 1246.970])
findCentre(nup107N_00_7r5k, nup107N_05_7r5k)

nup107N_00_dilated = np.array([520.991, 919.384, 1252.135])
nup107N_05_dilated = np.array([1466.209, 1067.816, 1252.135])

findCentre(nup107N_00_dilated, nup107N_05_dilated)
